#!/usr/bin/env python3
"""
Create a sphinx build environment for the OpenTURNS documentation
with only some of the source files in order to accelerate Sphinx builds.
Sphinx is then called to build the HTML doc using all available cores.
HTML files are written to a subfolder of the build folder called `install`.

Parameters
----------
build_name : str
    Name of the build folder

source_files : sequence of str
    Names of all source example Python scripts and RST source files to be included.

Notes
-----

Depends on the following Python libraries:

  * python3.8+
  * openturns
  * matplotlib
  * sphinx
  * sphinx-gallery
  * numpydoc
  * sphinx-copybutton
  * sphinxcontrib-jquery

Sphinx reads the docstrings of the OpenTURNS methods to generate the API doc.
This means that changing the source _doc.i.in files will have no effect,
since those files are not read by Sphinx.

The script attempts to address this problem by relying on the git diff command
to find out which parts of the source _doc.i.in files were changed and patch
the docstrings in the Python files of the OpenTURNS install folder accordingly.
The patch is reversed at the end of the script to restore the OpenTURNS install
folder to its original state.

By default, the script uses the command `git diff python/src/*_doc.i.in`
(which takes into account unstaged changes only),
but an additional argument xxx can be passed using the --diff option:

$ utils/docfast.py --diff xxx ...

will result in `git diff xxx python.src/*_doc.i.in` being used instead.

Be aware that the patch can fail in various ways. That is why the --nodiff
option is available to suppress the attempt to guess changes to the _doc.i.in
source files. That way, the OpenTURNS install folder is left untouched:

$ utils/docfast.py --nodiff ...

Examples
--------

Create a build directory `build-sphinxonly` that will contain only source files
required by Sphinx.

$ utils/docfast.py build-sphinxonly

Create a build directory `build-sphinxonly` that will contain only the mandatory
source files plus the `plot_smoothing_mixture.py` and `base_objects.rst` files.

$ utils/docfast.py build-sphinxonly base_objects.rst plot_smoothing_mixture.py

Do the same, but apply the changes made to the source _doc.i.in files with
respect to the master branch.

$ utils/docfast.py build-sphinxonly base_objects.rst plot_smoothing_mixture.py --diff master

"""
import os
import sys
import shutil
import glob

# from sphinx.cmd import build
import subprocess
import openturns as ot


def create_patch(gitdiff_arg):
    """
    Attempt to translate differences in _doc.i.in files into differences in the actual docstrings.

    Parameters
    ----------
    gitdiff_arg : str
        Additional argument to be passed to git diff
    """
    script_folder = os.path.dirname(os.path.abspath(__file__))  # openturns/utils
    sources_folder = os.path.dirname(script_folder)  # root folder openturns
    src_folder = os.path.join(sources_folder, "python", "src")  # openturns/python/src
    docstring_files = os.path.join(src_folder, "*_doc.i.in")
    gitdiff_filename = os.path.join(
        script_folder, "changes.patch"
    )  # openturns/utils/changes.patch

    # Call git diff
    if gitdiff_arg is None:
        gitdiff_listargs = ["git", "diff", docstring_files]
    else:
        gitdiff_listargs = ["git", "diff", gitdiff_arg, docstring_files]
    gitdiff = subprocess.check_output(gitdiff_listargs).decode()

    # Divide the output of git diff into chunks
    # Each chunks contains changes to one _doc.i.in file
    chunkstart = "diff --git a/"
    chunks = [chunkstart + chunk for chunk in gitdiff.split(chunkstart)[1:]]

    # Parse the chunks and try to recreate the actual docstring context
    for chunk_number, chunk in enumerate(chunks):
        lines = chunk.split("\n")
        for line_number, line in enumerate(lines[5:]):
            if len(line) == 0:
                lines[5 + line_number] = line
            elif line[0] == "@":
                lines[5 + line_number] = line
            elif line[0] == "+":
                if len(line) == 1:
                    lines[5 + line_number] = line
                elif line[1] == '"':
                    lines[5 + line_number] = "+" + "    " + line[2:]
                else:
                    lines[5 + line_number] = "+" + "    " + line[1:]
            elif line[0] == "-":
                if len(line) == 1:
                    lines[5 + line_number] = line
                elif line[1] == '"':
                    lines[5 + line_number] = "    " + '"""\n' + "-" + "    " + line[2:]
                else:
                    lines[5 + line_number] = "-" + "    " + line[1:]
            elif line[0] == '"':
                lines[5 + line_number] = "    " + '"""\n' + line
            else:
                lines[5 + line_number] = "    " + line
            if len(line) > 0:
                if lines[5 + line_number][-1] == '"':
                    if line[0] == "-":
                        lines[5 + line_number] = lines[5 + line_number][0:-1]
                    else:
                        lines[5 + line_number] = (
                            lines[5 + line_number][0:-1] + "\n" + "    " + '"""'
                        )
        lines = [line + "\n" for line in lines]
        chunks[chunk_number] = "".join(lines)

    # Join the chunks into one single string containing all diffs
    gitdiff = "".join(chunks)

    # Write the doctored string to changes.patch
    gitdiff_file = open(gitdiff_filename, "w")
    gitdiff_file.write(gitdiff)
    gitdiff_file.close()


def apply_patch():
    """
    Apply the patch to the Python files of the install folder.
    Create backup files to be used at the end of the script in order
    to rollback the patch.
    """
    script_folder = os.path.dirname(os.path.abspath(__file__))  # openturns/utils
    ot_folder = os.path.dirname(os.path.abspath(ot.__file__))  # install folder
    gitdiff_filename = os.path.join(
        script_folder, "changes.patch"
    )  # openturns/utils/changes.patch

    # Create backup Python scripts, apply the patch, remove byproducts
    for pypath in glob.iglob(os.path.join(ot_folder, "**", "*.py"), recursive=True):
        shutil.copyfile(pypath, pypath[0:-3] + ".BACKUP")
        subprocess.run(["patch", "-u", "-l", pypath, "-i", gitdiff_filename])
        try:
            os.remove(pypath + ".rej")
        except FileNotFoundError:
            pass
        try:
            os.remove(pypath + ".orig")
        except FileNotFoundError:
            pass

    # Remove patch file
    os.remove(gitdiff_filename)


# Get the arguments
arguments = sys.argv[1:]

if "--diff" in arguments and "--nodiff" in arguments:
    raise ValueError("--diff and --nodiff are incompatible")

# Options --diff and --nodiff determine gitdiff_flag and gitdiff_arg
gitdiff_flag = True
gitdiff_arg = None

# Non-optional arguments determine build_folder_name and list_preserved_files
build_folder_name = None
list_preserved_files = []

# Parse arguments
index = 0
while index < len(arguments):
    if arguments[index] == "--diff":
        if index + 1 == len(arguments):
            raise ValueError("No argument given after --diff")
        else:
            gitdiff_arg = arguments[index + 1]
            index += 1
    elif arguments[index] == "--nodiff":
        gitdiff_flag = False
    elif arguments[index][0] == "-":
        raise ValueError("Unrecognized option {}".format(arguments[index]))
    else:
        if build_folder_name is None:
            build_folder_name = arguments[index]
        else:
            list_preserved_files.append(arguments[index])
    index += 1

# Check arguments
if build_folder_name is None:
    raise ValueError("Build folder name should be passed as argument.")

# Create and apply patch to OT install Python files if required
if gitdiff_flag:
    create_patch(gitdiff_arg)
    apply_patch()

script_folder = os.path.dirname(os.path.abspath(__file__))  # openturns/utils
sources_folder = os.path.dirname(script_folder)  # root folder openturns
doc_folder = os.path.join(sources_folder, "python", "doc")  # openturns/python/doc
build_folder = os.path.join(sources_folder, build_folder_name)

# Copy everything from python/doc into the build folder
shutil.copytree(doc_folder, build_folder, dirs_exist_ok=True)  # Python 3.8+

# Remove all Python scripts in the examples folder...
for pypath in glob.iglob(
    os.path.join(build_folder, "examples", "**", "*.py"), recursive=True
):
    # ...unless their name was passed as argument.
    if os.path.basename(pypath) not in list_preserved_files:
        os.remove(pypath)

list_rstsections = ["developer_guide", "theory", "usecases", "user_manual"]

# For every doc section listed...
for section in list_rstsections:
    # ...remove all RST files...
    for rstpath in glob.iglob(
        os.path.join(build_folder, section, "**", "*.rst"), recursive=True
    ):
        # ...that do not have the same name as their directory...
        if (
            os.path.basename(rstpath)
            != os.path.basename(os.path.dirname(rstpath)) + ".rst"
        ):
            # ...unless their name was passed as argument.
            if os.path.basename(rstpath) not in list_preserved_files:
                os.remove(rstpath)

# Rename conf.py.in -> conf.py
os.rename(
    os.path.join(build_folder, "conf.py.in"), os.path.join(build_folder, "conf.py")
)

# build.build_main(['-j', 'auto', build_folder, os.path.join(build_folder, 'install')]) # does not work when git diff is used (because it shares a Python session?)
subprocess.run(
    ["sphinx-build", "-j", "auto", build_folder, os.path.join(build_folder, "install")]
)

# Restore all python files in openturns install directory
ot_folder = os.path.dirname(os.path.abspath(ot.__file__))  # install folder
for pypath in glob.iglob(os.path.join(ot_folder, "**", "*.BACKUP"), recursive=True):
    os.remove(pypath[0:-7] + ".py")
    os.rename(pypath, pypath[0:-7] + ".py")



from sphinx_gallery.scrapers import matplotlib_scraper
import matplotlib as mpl

mpl.rcParams["svg.hashsalt"] = "nacl"


def matplotlib_svg_scraper(*args, **kwargs):
    return matplotlib_scraper(*args, format="svg", **kwargs)



"""
A quick start guide to the `Point` and `Sample` classes
=======================================================
"""

# %%
# Abstract
# --------
#
# In this example, we present the :class:`~openturns.Point` and :class:`~openturns.Sample` classes, two fundamental objects in the library.
# We present the principles behind these classes and the way to create and use these objects.
# We show how to extract a row or a column with the slicing operator.
# We show how these objects interact with Python variables and with the Numpy module.

# %%
# Introduction
# ------------
#
# Two fundamental objects in the library are:
#
# * :class:`~openturns.Point`: a multidimensional point in :math:`d` dimensions (:math:`\in \Rset^d`) ;
# * :class:`~openturns.Sample`: a multivariate sample made of :math:`n` points in :math:`d` dimensions.
#

# %%
import numpy as np
import openturns as ot


# %%
# The `Point` class
# -----------------
#
# In this section, we see how to:
#
# * create a point in :math:`\mathbb{R}^3`,
# * access its components,
# * update its components.
#

# %%
# By default, points are filled with zeros.

# %%
p = ot.Point(3)
p

# %%
# The following statement returns the value of the second component (with index 1).
# Python beginners should remember that Python indices start at zero.

# %%
p[1]

# %%
# The following statement sets the second component.

# %%
p[1] = 2
p

# %%
p.getDimension()

# %%
# The `Sample` class
# ------------------
#
# The :class:`~openturns.Sample` class represents a multivariate sample made of :math:`\sampleSize` points in :math:`\Rset^d`.
#
# * :math:`d` is the *dimension* of the sample,
# * :math:`n` is the *size* of the sample.
#
#
# A `Sample` can be seen as an array of with :math:`n` rows and :math:`d` columns.
#
# *Remark.* The :class:`~openturns.ProcessSample` class can be used to manage a sample of stochastic processes.

# %%
# The script below creates a `Sample` with size :math:`n=5` and dimension :math:`d=3`.

# %%
data = ot.Sample(5, 3)
data

# %%
data.getSize()

# %%
data.getDimension()

# %%
# The following statement sets the third component (with index 2) of the fourth point (with index 3) in the `Sample`.

# %%
data[3, 2] = 32
data

# %%
# Notice that the rendering is different when we use the `print` statement.

# %%
print(data)

# %%
# We can customize the format used to print the floating point numbers
# with the `Sample-PrintFormat` key of the :class:`~openturns.ResourceMap`.

# %%
# Get a row or a column of a `Sample`
# -----------------------------------
#
# As with `numpy` arrays, we can extract a row or a column with the `:` slicing operator.
# As a reminder for Python beginners, *slicing* is the fact of extracting a part of an array with one single statement; this avoids `for` loops and improves performance and readability.

# %%
row = data[3, :]
row

# %%
print(type(row))

# %%
column = data[:, 2]
column

# %%
print(type(column))

# %%
# We see that:
#
# * the `row` is a `Point`,
# * the `column` is a `Sample`.
#
# This is consistent with the fact that, in a dimension :math:`d`, a row is a :math:`d`-dimensional `Point`.

# %%
# The following statement extracts several columns (with indices 0 and 2) and creates a new `Sample`.

# %%
data.getMarginal([0, 2])

# %%
# Set a row or a column of a `Sample`
# -----------------------------------

# %%
# Slicing can also be used to set a `Sample` row or column.

# %%
sample = ot.Sample([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
sample

# %%
# Set the third row: this must be a `Point` or must be convertible to.
p = [8.0, 10.0]
sample[2, :] = p
sample

# %%
# Set the second column: this must be a `Sample` or must be convertible to.
sample = ot.Sample([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
s = ot.Sample([[3.0], [5.0], [7.0]])
sample[:, 1] = s
sample

# %%
# Sometimes, we want to set a column with a list of floats.
# This can be done using the :meth:`~openturns.Sample.BuildFromPoint` static method.
sample = ot.Sample([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
s = ot.Sample.BuildFromPoint([3.0, 5.0, 7.0])
sample[:, 1] = s
sample

# %%
# Create a `Point` or a `Sample` from a Python list
# -------------------------------------------------
#
# The following statement creates a `Point` from a Python list.

# %%
p1 = ot.Point([2, 3])
p1

# %%
p2 = ot.Point(range(2))
p2

# %%
# The first useful *Pythonism* that we will review is the *list comprehension*. This creates a list from a `for` loop.
# This kind of statements is often used in the examples, so that they can be as short as possible.
# In the following statement, we create a point by iterating over the components of a `Point`.

# %%
p3 = ot.Point([i * i for i in p1])
p3

# %%
# The second useful *Pythonism* is the repetition with the `*` operator.
#
# The following statements creates a list with three 5s.

# %%
p4 = [5] * 3
p4

# %%
# We can also create a `Sample` from a list of `Point`.

# %%
sample = ot.Sample([p1, p2, p3])
sample

# %%
# We can loop over the points in a sample, using a list comprehension. In the following example, we compute the Euclidian norm of the points in the previous sample.

# %%
[point.norm() for point in sample]

# %%
# We can also create a `Sample` based on a `Point`, repeated three times.

# %%
sample = ot.Sample([p4] * 3)
sample

# %%
# A nested list of floats is the easiest way to create a non-trivial `Sample`.

# %%
sample = ot.Sample([[0, 1], [2, 3], [4, 5]])
sample

# %%
# Extract some specified points from a `Sample`
# ---------------------------------------------
#
# We can extract some points from a Sample. There are several ways to do this.
#
# Let us create a Sample of dimension 3 and size 10.

# %%
size = 6
sample = ot.Normal(6).getSample(size)
sample

# %%
# We can extract a point which index is explicitly known. For example, to get the point indexed by 0, we can write:
point1 = sample[0]
point1

# %%
# To get the points indexed by 1 and 3, we can write:
indices = [1, 3]
selected_points = sample[indices]
selected_points

# %%
# or:
selected_points = sample.select(indices)
selected_points


# %%
# We can also select the indices with a rule. For example, we want to extract all the points which third component is negative.
selected_points = sample[[i for i in range(sample.getSize()) if sample[i, 2] < 0.0]]
selected_points

# %%
# or:
selected_points = sample.select(
    [i for i in range(sample.getSize()) if sample[i, 2] < 0.0]
)
selected_points

# %%
# We can also extract some specified points reduced to some specified components. For example,
# to get the points indexed by 0 and 1, and reduced to the components
# indexed by 4 and 5, we can write:
selected_points = sample[0:2, [4, 5]]
selected_points

# %%
# or:
selected_points = sample[0:2, 4:6]
selected_points

# %%
# Interactions with Numpy
# -----------------------
#
# Classes defined in pure Python modules cannot be used by the library.
# This is why it is useful to know how to convert to and from more basic Python variable types, especially Numpy arrays.

# %%
# The following statement creates a :class:`~openturns.Sample` and converts it into a bidimensional Numpy `array`.

# %%
sample = ot.Sample(5, 3)
array = np.array(sample)
array

# %%
print(type(array))

# %%
# Conversely, the following script creates a Numpy `array`, then converts it into a :class:`~openturns.Sample`.

# %%
array = 3.14 * np.ones((5, 3))
sample = ot.Sample(array)
sample

# %%
sample.getSize()

# %%
sample.getDimension()

# %%
# There is an ambiguous situation: a `Sample` based on several scalar values.
#
# For example, is a `Sample` based on 5 values:
#
# * a `Sample` with size 5 in 1 dimension or
# * a `Sample` with size 1 in 5 dimensions?
#
# In order to solve the case, we can use the second input argument of the `Sample` constructor, which specifies the dimension.
#
# The following statement creates an array containing 5 values from 0 to 1.

# %%
u = np.linspace(0, 1, 5)
u

# %%
# Choice A: we create a `Sample` with size 5 in 1 dimension.

# %%
sample = ot.Sample([[ui] for ui in u])
sample

# %%
# Choice B: we create a `Sample` with size 1 in 5 dimensions.

# %%
sample = ot.Sample([u[i : i + 5] for i in range(len(u) // 5)])
sample

# %%
# When there is an ambiguous case, the library cannot solve the
# issue and an `InvalidArgumentException` is generated.

# %%
# More precisely, the code:
#
# .. code-block::
#
#     sample = ot.Sample(u)
#

# %%
# produces the exception:
#
# .. code-block::
#
#     TypeError: InvalidArgumentException : Invalid array dimension: 1
#

# %%
# In order to solve that problem, we can use the :meth:`~openturns.Sample.BuildFromPoint`
# static method.
sample = ot.Sample.BuildFromPoint([ui for ui in u])
sample



#! /usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

rv = ot.RandomVector()

# load
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyrv.xml"))
study.load()
study.fillObject("rv", rv)

print(rv)
print(rv.getRealization())
print(rv.getMean())

os.remove("pyrv.xml")



import openturns as ot
from math import exp
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Create the covariance function at (s,t)


def C(s, t):
    return exp(-4.0 * abs(s - t) / (1 + (s * s + t * t)))


def covModRef(X):
    return [C(X[0], X[1])]


myFuncCovarianceRef = ot.PythonFunction(2, 1, covModRef)
myFuncCovarianceRef.setDescription(["s", "t", "C"])

t0 = -4.0
tmax = 4.0

# Draw the isocontours of the discretized covariance function
myGraphRef = myFuncCovarianceRef.draw([t0, t0], [tmax, tmax])
alld = myGraphRef.getDrawables()
levels = ot.Point(alld.getSize())
for i in range(alld.getSize()):
    d = alld[i]
    d.setLineStyle("twodash")
    d.setLineWidth(2)
    myGraphRef.setDrawable(i, d)
    levels[i] = d.getLevels()[0]


# Create the time grid
# for iN in range(2, 11):
#  t_00 = time()
N = 2**5
dt = (tmax - t0) / N
myMesh = ot.RegularGrid(t0, dt, N)

# Keep only time stamps in the time-grid
tmax = myMesh.getEnd()

# Create the collection of HermitianMatrix
covariance = ot.CovarianceMatrix(N)
for k in range(N):
    s = myMesh.getValue(k)
    for ll in range(k + 1):
        t = myMesh.getValue(ll)
        covariance[k, ll] = C(s, t)

# Create the covariance model
myCovarianceModel = ot.UserDefinedCovarianceModel(myMesh, covariance)

# Create the non stationary Gaussian process with
# that covariance model
myProcess = ot.GaussianProcess(myCovarianceModel, myMesh)

# Create a  sample of fields
size = 10**4
myFieldSample = myProcess.getSample(size)

# Build a covariance model factory
myFactory = ot.NonStationaryCovarianceModelFactory()

# Estimation on a the ProcessSample
myEstimatedModel = myFactory.build(myFieldSample)

# Define the python function associated to myCovarianceModel


def covMod(X):
    return [myEstimatedModel(X[0], X[1])[0, 0]]


myFuncCovariance = ot.PythonFunction(2, 1, covMod)


cov_graph = ot.Graph(myGraphRef)
alld = myFuncCovariance.draw([t0, t0], [tmax, tmax]).getDrawables()
palette = ot.Drawable.BuildDefaultPalette(alld.getSize())
for i in range(alld.getSize()):
    d = alld[i]
    d.setLegend("")
    d.setLevels([levels[i]])
    d.setColor(palette[i])
    d.setDrawLabels(False)
    cov_graph.add(d)
cov_graph.setTitle("Non stationary covariance model estimation")

fig = plt.figure(figsize=(10, 4))
cov_axis = fig.add_subplot(111)
otv.View(cov_graph, figure=fig, axes=[cov_axis], add_legend=False, square_axes=True)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Product
# First, build a function from R^3->R
inVar = ["x0", "x1", "x2"]
formula = ["x0^2 + 2 * x1 * x2 + 3 * x2"]
f1 = ot.SymbolicFunction(inVar, formula)
# Second, build a function from R^3->R^2
formula = ["x2 - x0 + x1"]
formula.append("x0 + x1 * x0 + x2")
f2 = ot.SymbolicFunction(inVar, formula)
# Third, build the product function
myFunction = f1 * f2
inPoint = ot.Point([1.2, 2.3, 3.4])
print("myFunction=", myFunction)
value = myFunction(inPoint)
print("Value at  %s =\n%s" % (inPoint, value))
gradient = myFunction.gradient(inPoint)
print("Gradient at  %s =\n%s" % (inPoint, gradient))
hessian = myFunction.hessian(inPoint)
print("Hessian at  %s =\n%s" % (inPoint, hessian))

# Sum/difference
# First, build two functions from R^3->R^2
inVar = ["x0", "x1", "x2"]
formula = ["x0 + 2 * x1 * x2 + 3 * x2", "x2 - x0 + x1 * x0"]
f1 = ot.SymbolicFunction(inVar, formula)
formula = ["x0 + x1 + x2", "-2 * x0 + 3 * x2 * x1 - x1"]
f2 = ot.SymbolicFunction(inVar, formula)
# Second, build the function
mySum = f1 + f2
print("mySum=", mySum)
value = mySum(inPoint)
print(f"Value at  {inPoint} =\n{value}")
gradient = mySum.gradient(inPoint)
print(f"Gradient at  {inPoint} =\n{gradient}")
hessian = mySum.hessian(inPoint)
print(f"Hessian at  {inPoint} =\n{hessian}")
myDiff = f1 - f2
print("myDiff=", myDiff)
value = myDiff(inPoint)
print(f"Value at  {inPoint} =\n{value}")
gradient = myDiff.gradient(inPoint)
print(f"Gradient at  {inPoint} =\n{gradient}")
hessian = myDiff.hessian(inPoint)
print(f"Hessian at  {inPoint} =\n{hessian}")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Database construction
dimension = 2

inputSample = ot.Sample(0, dimension)
inputSample.add(ot.Point(dimension, 1.0))
inputSample.add(ot.Point(dimension, 2.0))
inputSample.setDescription(["x0", "x1"])
outputSample = ot.Sample(0, 1)
outputSample.add(ot.Point(1, 4.0))
outputSample.add(ot.Point(1, 5.0))
outputSample.setDescription(["y0"])
database = ot.DatabaseFunction(inputSample, outputSample)

print("database=", database)

# Does it work?
x = ot.Point(dimension, 1.8)
print("x=", x)
print("database(x)=", database(x))



"""
Estimate a GPD on the daily rainfall data
=========================================
"""

# %%
# In this example, we illustrate various techniques of extreme value modeling applied
# to the daily rainfall accumulations in south-west England, over the period 1914-1962.
# Readers should refer to [coles2001]_ to get more details.
#
# We illustrate techniques to:
#
# - estimate a stationary and a non stationary GPD,
# - estimate a return level,
#
# using:
#
# - the log-likelihood function,
# - the profile log-likelihood function.
#
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import coles

# %%
# First, we load the Rain dataset. We start by looking at it through time.
dataRain = coles.Coles().rain
print(dataRain[:10])
graph = ot.Graph(
    "Daily rainfall accumulations SW England", "day", "level (mm)", True, ""
)
days = ot.Sample([[i] for i in range(len(dataRain))])
cloud = ot.Cloud(days, dataRain)
cloud.setColor("red")
cloud.setPointStyle(",")
graph.add(cloud)
graph.setIntegerXTick(True)
view = otv.View(graph)

# %%
# In order to select a threshold upon which the GPD model will be fitted, we draw
# the mean residual life plot with approximate :math:`95\%` confidence interval.
# It appears that the graph is linear from the threshold around
# :math:`u=30`. Then, it decays sharply although with a linear trend. We
# should be tempted to choose :math:`u=60` but there are only 6
# exceedances of the threshold :math:`u=60`. So it is not enough
# to make meaningful inferences. Moreover, the graph is not reliable
# for large values of :math:`u` due to the limited amount of data on
# which the estimate and the confidence interval are based.
# For all these reasons, it appears preferable to chose :math:`u=30`.
factory = ot.GeneralizedParetoFactory()
graph = factory.drawMeanResidualLife(dataRain)
view = otv.View(graph)

# %%
# To support that choice, we draw the parameter stability plots.
# We see that the perturbations appear small relatively to sampling errors.
# We can see that the change in pattern observed in the mean
# residual life plot is still apparent here for high thresholds.
# Hence, we choose the threshold :math:`u=30`.
u_range = ot.Interval(0.5, 50.0)
graph = factory.drawParameterThresholdStability(dataRain, u_range)
view = otv.View(graph, figure_kw={"figsize": (6.0, 6.0)})

# %%
# **Stationary GPD modeling via the log-likelihood function**
#
# We first assume that the dependence through time is negligible, so we first model the data as
# independent observations over the observation period.
#
# We consider the model  :math:`\mathcal{M}_0` defined by:
#
# .. math::
#     :nowrap:
#
#     \begin{align*}
#       \sigma(t) & = \sigma\\
#       \xi(t) & = \xi
#     \end{align*}
#
# We estimate the parameters of the GPD distribution by maximizing
# the log-likelihood of the data for the selecte threshold :math:`u=30`.
u = 30
result_LL = factory.buildMethodOfLikelihoodMaximizationEstimator(dataRain, u)

# %%
# We get the fitted GPD and its parameters of :math:`(\hat{\sigma}, \hat{\xi}, u)`.
fitted_GPD = result_LL.getDistribution()
desc = fitted_GPD.getParameterDescription()
param = fitted_GPD.getParameter()
print(", ".join([f"{p}: {value:.3f}" for p, value in zip(desc, param)]))
print("log-likelihood = ", result_LL.getLogLikelihood())

# %%
# We get the asymptotic distribution of the estimator :math:`(\hat{\mu}, \hat{\sigma}, \hat{\xi})`.
# In that case, the asymptotic distribution is normal.
parameterEstimate = result_LL.getParameterDistribution()
print("Asymptotic distribution of the estimator : ")
print(parameterEstimate)

# %%
# We get the covariance matrix and the standard deviation of :math:`(\hat{\sigma}, \hat{\xi}, \hat{\xi})`.
print("Cov matrix = \n", parameterEstimate.getCovariance())
print("Standard dev = ", parameterEstimate.getStandardDeviation())

# %%
# We get the marginal confidence intervals of order 0.95.
order = 0.95
for i in range(2):  # exclude u parameter (fixed)
    ci = parameterEstimate.getMarginal(i).computeBilateralConfidenceInterval(order)
    print(desc[i] + ":", ci)

# %%
# At last, we can validate the inference result thanks to the 4 usual diagnostic plots:
#
# - the probability-probability pot,
# - the quantile-quantile pot,
# - the return level plot,
# - the data histogram and the density of the fitted model.
#
validation = ot.GeneralizedParetoValidation(result_LL, dataRain)
graph = validation.drawDiagnosticPlot()
view = otv.View(graph)

# %%
# **Stationary GPD modeling via the profile log-likelihood function**
#
# Now, we use the profile log-likehood function with respect
# to the :math:`\xi` parameter rather than log-likehood function
# to estimate the parameters of the GPD.
result_PLL = factory.buildMethodOfXiProfileLikelihoodEstimator(dataRain, u)

# %%
# The following graph allows one to get the profile log-likelihood plot.
# It also indicates the optimal value of :math:`\xi`, the maximum profile log-likelihood and
# the confidence interval for :math:`\xi` of order 0.95 (which is the default value).
order = 0.95
result_PLL.setConfidenceLevel(order)
view = otv.View(result_PLL.drawProfileLikelihoodFunction())

# %%
# We can get the numerical values of the confidence interval: it appears to be a bit smaller
# with the interval obtained from the profile log-likelihood function than with the log-likelihood
# function.
# Note that if the order requested is too high, the confidence interval might not be calculated because
# one of its bound is out of the definition domain of the log-likelihood function.
try:
    print("Confidence interval for xi = ", result_PLL.getParameterConfidenceInterval())
except Exception as ex:
    print(type(ex))
    pass

# %%
# **Return level estimate from the estimated stationary GPD**
#
# We evaluate the :math:`T`-year return level which corresponds to the
# :math:`m`-observation return level, where :math:`m = T*n_y` with :math:`n_y`
# the number of observations per year. Here, we have daily observations, hence
# :math:`n_y = 365`. As we assumed that the observations were independent, the extremal index is :math:`\theta=1` which is the default value.
#
# The method also provides the asymptotic distribution of the estimator :math:`\hat{z}_m`.
ny = 365
T10 = 10
zm_10 = factory.buildReturnLevelEstimator(result_LL, dataRain, T10 * ny)
return_level_10 = zm_10.getMean()
print("Maximum log-likelihood function : ")
print(f"10-year return level = {return_level_10}")
return_level_ci10 = zm_10.computeBilateralConfidenceInterval(0.95)
print(f"CI = {return_level_ci10}")

T100 = 100
zm_100 = factory.buildReturnLevelEstimator(result_LL, dataRain, T100 * ny)
return_level_100 = zm_100.getMean()
print(f"100-year return level = {return_level_100}")
return_level_ci100 = zm_100.computeBilateralConfidenceInterval(0.95)
print(f"CI = {return_level_ci100}")

# %%
# **Return level estimate via the profile log-likelihood function of a stationary GPD**
#
# We can estimate the :math:`m`-observation return level :math:`z_m` directly from the data using the profile
# likelihood with respect to :math:`z_m`.
result_zm_100_PLL = factory.buildReturnLevelProfileLikelihoodEstimator(
    dataRain, u, T100 * ny
)
zm_100_PLL = result_zm_100_PLL.getParameter()
print(f"100-year return level (profile) = {zm_100_PLL}")

# %%
# We can get the confidence interval of :math:`z_m`: once more, it appears to be a bit smaller
# than the interval obtained from the log-likelihood function.
result_zm_100_PLL.setConfidenceLevel(0.95)
return_level_ci100 = result_zm_100_PLL.getParameterConfidenceInterval()
print("Maximum profile log-likelihood function : ")
print(f"CI={return_level_ci100}")

# %%
# We can also plot the profile log-likelihood function and get the confidence interval, the optimal value
# of :math:`z_m` and its confidence interval.
view = otv.View(result_zm_100_PLL.drawProfileLikelihoodFunction())

# %%
# **Non stationary GPD modeling via the log-likelihood function**
#
# Now, we want to check whether it is necessary to model the time dependency over
# the observation period.
#
# We have to define the functional basis for each parameter of the GPD model. Even if we have
# the possibility to affect a time-varying model to each of the 2 parameters
# :math:`(\sigma, \xi)`,
# it is strongly recommended not to vary the shape parameter :math:`\xi`.
#
# We suppose that :math:`\sigma` is linear with time, and that the other parameters remain constant.
#
# For numerical reasons, it is strongly recommended to normalize all the data as follows:
#
# .. math::
#
#     \tau(t) = \dfrac{t-c}{d}
#
# where:
#
# - the *CenterReduce* method where :math:`c = \dfrac{1}{n} \sum_{i=1}^n t_i` is the mean time stamps
#   and :math:`d = \sqrt{\dfrac{1}{n} \sum_{i=1}^n (t_i-c)^2}` is the standard deviation of the time stamps;
# - the *MinMax* method where :math:`c = t_1` is the initial time and :math:`d = t_n-t_1` the final time. This method is the default one;
# - the *None* method where :math:`c = 0` and :math:`d = 1`: in that case, data are not normalized.
#
# We consider the model  :math:`\mathcal{M}_1` defined by:
#
# .. math::
#     :nowrap:
#
#     \begin{align*}
#       \sigma(t) & = \beta_1 + \beta_2\tau(t) \\
#       \xi(t) & = \beta_3
#     \end{align*}
#
constant = ot.SymbolicFunction(["t"], ["1.0"])
basis = ot.Basis([ot.SymbolicFunction(["t"], ["t"]), constant])
# basis for mu, sigma, xi
sigmaIndices = [0, 1]  # linear
xiIndices = [1]  # stationary

# %%
# We need to get the time stamps (in days here).
timeStamps = ot.Sample([[i + 1] for i in range(len(dataRain))])

# %%
# We can now estimate the list of coefficients :math:`\vect{\beta} = (\beta_1, \beta_2, \beta_3)` using
# the log-likelihood of the data.
result_NonStatLL = factory.buildTimeVarying(
    dataRain, u, timeStamps, basis, sigmaIndices, xiIndices
)
beta = result_NonStatLL.getOptimalParameter()
print(f"beta = {beta}")
print(f"sigma(t) = {beta[1]:.4f} * tau(t) + {beta[0]:.4f}")
print(f"xi = {beta[2]:.4f}")
print(f"Max log-likelihood = {result_NonStatLL.getLogLikelihood()}")

# %%
# We get the asymptotic distribution of :math:`\vect{\beta}` to compute some confidence intervals of
# the estimates, for example of order :math:`p = 0.95`.
dist_beta = result_NonStatLL.getParameterDistribution()
confidence_level = 0.95
for i in range(beta.getSize()):
    lower_bound = dist_beta.getMarginal(i).computeQuantile((1 - confidence_level) / 2)[
        0
    ]
    upper_bound = dist_beta.getMarginal(i).computeQuantile((1 + confidence_level) / 2)[
        0
    ]
    print(
        "Conf interval for beta_"
        + str(i + 1)
        + " = ["
        + str(lower_bound)
        + "; "
        + str(upper_bound)
        + "]"
    )

# %%
# You can get the expression of the normalizing function :math:`t \mapsto \tau(t)`:
normFunc = result_NonStatLL.getNormalizationFunction()
print("Function tau(t): ", normFunc)
print("c = ", normFunc.getEvaluation().getImplementation().getCenter()[0])
print("1/d = ", normFunc.getEvaluation().getImplementation().getLinear()[0, 0])

# %%
# You can get the function :math:`t \mapsto \vect{\theta}(t)` where
# :math:`\vect{\theta}(t) = (\sigma(t), \xi(t))`.
functionTheta = result_NonStatLL.getParameterFunction()

# %%
# In order to compare different modelings, we get the optimal log-likelihood of the data for both stationary
# and non stationary models. The difference seems to be non significant enough, which means that the non
# stationary model does not really improve the quality of the modeling.
print("Max log-likelihood: ")
print("Stationary model =  ", result_LL.getLogLikelihood())
print("Non stationary linear sigma(t) model =  ", result_NonStatLL.getLogLikelihood())

# %%
# In order to draw some diagnostic plots similar to those drawn in the stationary case, we refer to the
# following result: if :math:`Y_t` is a non stationary GPD model parametrized by :math:`(\sigma(t), \xi(t), u)`,
# then the standardized variables :math:`\hat{Y}_t` defined by:
#
# .. math::
#
#    \hat{Y}_t = \dfrac{1}{\xi(t)} \log \left[1+ \xi(t)\left( \dfrac{Y_t-u}{\sigma(t)} \right)\right]
#
# have the Exponential distribution which is the GPD model with :math:`(\sigma, \xi, u) = (1, 0, 0)`.
#
# As a result, we can validate the inference result thanks the 4 usual diagnostic plots:
#
# - the probability-probability pot,
# - the quantile-quantile pot,
# - the return level plot,
# - the data histogram and the density of the fitted model.
#
# using the transformed data compared to the Exponential model. We can see that the adequation seems similar to the graph
# of the stationary model.
graph = result_NonStatLL.drawDiagnosticPlot()
view = otv.View(graph)

# %%
# We can draw the mean function :math:`t \mapsto \Expect{\mbox{GPD}(t)}`, defined for :math:`\xi <1` only:
#
# .. math::
#     \Expect{\mbox{GPD}(t)} = u + \dfrac{\sigma(t)}{1 - \xi(t)}
#
# We can also draw the function :math:`t \mapsto q_p(t)` where :math:`q_p(t)` is the quantile of
# order :math:`p` of the GPD distribution at time :math:`t`.
# Here, :math:`\sigma(t)` is a linear function and the other parameters are constant, so the mean and the quantile
# functions are also linear functions.
graph = ot.Graph(
    r"Maximum rain - Linear $\sigma(t)$",
    "day",
    "level (mm)",
    True,
    "",
)
graph.setIntegerXTick(True)
# data
cloud = ot.Cloud(timeStamps, dataRain)
cloud.setColor("red")
graph.add(cloud)
# mean function
meandata = [
    result_NonStatLL.getDistribution(t).getMean()[0] for t in timeStamps.asPoint()
]
curve_meanPoints = ot.Curve(timeStamps.asPoint(), meandata)
graph.add(curve_meanPoints)
# quantile function
graphQuantile = result_NonStatLL.drawQuantileFunction(0.95)
drawQuant = graphQuantile.getDrawable(0)
drawQuant = graphQuantile.getDrawable(0)
drawQuant.setLineStyle("dashed")
graph.add(drawQuant)
graph.setLegends(["data", "mean function", "quantile 0.95 function"])
graph.setLegendPosition("lower right")
view = otv.View(graph)

# %%
# At last, we can test the validity of the stationary model :math:`\mathcal{M}_0`
# relative to the model with time varying parameters  :math:`\mathcal{M}_1`. The
# model :math:`\mathcal{M}_0` is parametrized
# by :math:`(\beta_1, \beta_3)` and the model :math:`\mathcal{M}_1` is parametrized
# by :math:`(\beta_1, \beta_2, \beta_3)`: so we have :math:`\mathcal{M}_0 \subset \mathcal{M}_1`.
#
# We use the Likelihood Ratio test. The null hypothesis is the stationary model :math:`\mathcal{M}_0`.
# The Type I error :math:`\alpha` is taken equal to 0.05.
#
# This test confirms that there is no evidence of a linear trend for :math:`\mu`.
llh_LL = result_LL.getLogLikelihood()
llh_NonStatLL = result_NonStatLL.getLogLikelihood()
modelM0_Nb_param = 2
modelM1_Nb_param = 3
resultLikRatioTest = ot.HypothesisTest.LikelihoodRatioTest(
    modelM0_Nb_param, llh_LL, modelM1_Nb_param, llh_NonStatLL, 0.05
)
accepted = resultLikRatioTest.getBinaryQualityMeasure()
print(
    f"Hypothesis H0 (stationary model) vs H1 (linear mu(t) model):  accepted ? = {accepted}"
)

# %%
# We detail the statistics of the Likelihood Ratio test: the deviance statistics
# :math:`\mathcal{D}_p` follows a :math:`\chi^2_1` distribution.
# The model :math:`\mathcal{M}_0` is rejected if the deviance statistics estimated on the data is greater than
# the threshold :math:`c_{\alpha}` or if the p-value is less than the Type I error  :math:`\alpha = 0.05`.
print(f"Dp={resultLikRatioTest.getStatistic():.2f}")
print(f"alpha={resultLikRatioTest.getThreshold():.2f}")
print(f"p-value={resultLikRatioTest.getPValue():.2f}")


# %%
# We can test a linear trend in the log-scale parameter for :math:`\sigma(t)`:
#
# .. math::
#     :nowrap:
#
#     \begin{align*}
#       \sigma(t) & = exp(\beta_1 + \beta_2\tau(t)) \\
#       \xi(t) & = \beta_3
#     \end{align*}
sigmaLink = ot.SymbolicFunction("x", "exp(x)")
result_NonStatLL_Link = factory.buildTimeVarying(
    dataRain, u, timeStamps, basis, sigmaIndices, xiIndices, sigmaLink
)
beta = result_NonStatLL_Link.getOptimalParameter()
print(f"beta = {beta}")
print(f"sigma(t) = exp({beta[1]:.4f} * tau(t) + {beta[0]:.4f})")
print(f"xi = {beta[2]:.4f}")
print(f"Max log-likelihood = {result_NonStatLL.getLogLikelihood()}")

# %
# The maximized log-likelihood we obtain with the log-linear model is very similar
# to the one we obtained with the linear model. Hence, there is no evidence of a time trend.
# We draw the diagnostic plots which are similar to the previous ones.
graph = result_NonStatLL_Link.drawDiagnosticPlot()
view = otv.View(graph)

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import time

ot.TESTPREAMBLE()

dim = 2
f = ot.SymbolicFunction(["x1", "x2"], ["x1^2+x2^2"])
x0 = [0.5] * dim
bounds = ot.Interval([-1.0, -10.0], [10.0, 1.0])

for minimization in [True, False]:
    for ineq in [False, True]:
        problem = ot.OptimizationProblem(f)
        problem.setBounds(bounds)
        problem.setMinimization(minimization)
        if ineq:
            h = ot.SymbolicFunction(["x1", "x2"], ["5-x1", "x2+5"])
            problem.setInequalityConstraint(h)
        names = ot.OptimizationAlgorithm.GetAlgorithmNames(problem)
        for name in names:
            if "global" in name:
                # slow
                continue
            algo = ot.OptimizationAlgorithm.GetByName(name)
            algo.setProblem(problem)
            algo.setMaximumConstraintError(1e-1)
            algo.setMaximumCallsNumber(1000)
            try:
                algo.setStartingPoint(x0)
            except Exception:
                # multi-obj
                continue
            print(f"minimization={minimization} ineq={ineq} algo={name}")
            algo.run()
            result = algo.getResult()
            x = result.getOptimalPoint()
            y = result.getOptimalValue()
            print(f"x^={x} y^={y}")
            if minimization:
                assert abs(y[0]) < 0.3
            else:
                ref = 25.0 if ineq else 100.0
                assert y[0] > ref


def _exec(x):
    if x[0] < 0.0 or x[0] > 1.0:
        raise ValueError(f"Point {x[0]} not in bounds.")
    c = 1.0 - x[0] ** 2
    return [c]


# check algorithm stays inside bounds
eps = 1e-14
ot.ResourceMap.SetAsScalar("CenteredFiniteDifferenceGradient-DefaultEpsilon", eps)
ot.ResourceMap.SetAsScalar("CenteredFiniteDifferenceHessian-DefaultEpsilon", eps)
# the tolerance prevents FD gradients from stepping outside of [0,1]
# but we still want to keep it as low as possible to detect slightly out of bounds points
bounds = ot.Interval([2.0 * eps], [1.0 - 2.0 * eps])
costFunction = ot.PythonFunction(1, 1, _exec)
problem = ot.OptimizationProblem(costFunction)
problem.setBounds(bounds)
for name in ot.OptimizationAlgorithm.GetAlgorithmNames():
    algo = ot.OptimizationAlgorithm.GetByName(name)
    algo.setMaximumConstraintError(0.0)
    try:
        algo.setProblem(problem)
        startingPoint = [0.5]
        algo.setStartingPoint(startingPoint)
    except Exception:
        # not supported
        continue
    algo.setMaximumCallsNumber(100)
    algo.setMaximumIterationNumber(100)
    print(f"{name}...")
    algo.run()
    assert algo.getResult().getStatus() == ot.OptimizationResult.SUCCESS
    print(f"{name}: OK")


def _exec(X):
    time.sleep(0.2)
    x1, x2 = X
    return [x1**2 + x2**2]


# check algorithm enforces time limit
ot.Log.Show(ot.Log.ALL)
bounds = ot.Interval([-10] * 2, [10] * 2)
costFunction = ot.PythonFunction(2, 1, _exec)
problem = ot.OptimizationProblem(costFunction)
problem.setMinimization(False)
problem.setBounds(bounds)
for name in ot.OptimizationAlgorithm.GetAlgorithmNames():
    if name in ot.Bonmin.GetAlgorithmNames():
        # interruption has to be forced but cannot recover
        continue
    if "AUGLAG" in name:
        # returns XTOL_REACHED
        continue
    if name == "Ipopt":
        # walltime criterion requires ipopt>=3.14
        continue
    algo = ot.OptimizationAlgorithm.GetByName(name)
    try:
        algo.setProblem(problem)
        startingPoint = [0.1] * 2
        algo.setStartingPoint(startingPoint)
    except Exception:
        # not supported
        continue
    algo.setMaximumIterationNumber(50)
    algo.setMaximumCallsNumber(100)
    algo.setMaximumTimeDuration(0.1)
    algo.setCheckStatus(False)
    print(f"{name}...")
    algo.run()
    status = algo.getResult().getStatus()
    msg = algo.getResult().getStatusMessage()
    calls = algo.getResult().getCallsNumber()
    print(f"{name}: {status} {msg} {calls}")
    assert status == ot.OptimizationResult.TIMEOUT, name

# infeasible problem
objective = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"], ["x1 + 2 * x2 - 3 * x3 + 4 * x4"]
)
inequality_constraint = ot.SymbolicFunction(["x1", "x2", "x3", "x4"], ["-1.0"])
dim = objective.getInputDimension()
bounds = ot.Interval([-3.0] * dim, [5.0] * dim)
problem = ot.OptimizationProblem(objective)
problem.setMinimization(True)
problem.setInequalityConstraint(inequality_constraint)
problem.setBounds(bounds)
for name in ot.OptimizationAlgorithm.GetAlgorithmNames():
    algo = ot.OptimizationAlgorithm.GetByName(name)
    algo.setCheckStatus(False)
    try:
        algo.setProblem(problem)
        startingPoint = [0.0] * dim
        algo.setStartingPoint(startingPoint)
    except Exception:
        # not supported
        continue
    print(f"{name}...")
    try:
        algo.run()
    except Exception:
        # no feasible point
        pass
    result = algo.getResult()
    status = algo.getResult().getStatus()
    msg = algo.getResult().getStatusMessage()
    calls = algo.getResult().getCallsNumber()
    print(f"{name}: {status} {msg} {calls}")
    assert len(result.getOptimalPoint()) == 0, "should not find point"
    assert status == ot.OptimizationResult.FAILURE, "should return FAILURE"

# stop immediately
rosenbrock = ot.SymbolicFunction(["x1", "x2"], ["(1-x1)^2+100*(x2-x1^2)^2"])
problem = ot.OptimizationProblem(rosenbrock)
for name in ot.OptimizationAlgorithm.GetAlgorithmNames():
    algo = ot.OptimizationAlgorithm.GetByName(name)
    try:
        algo.setProblem(problem)
        algo.setStartingPoint([0.0] * 2)
    except Exception:
        # not supported
        continue
    print(f"{name}...")
    algo.setMaximumResidualError(1e-3)
    algo.setMaximumCallsNumber(10000)
    if name in list(ot.Bonmin.GetAlgorithmNames()) + ["Ipopt"]:
        algo.setCheckStatus(False)  # optim fails with "Invalid number detected"

    def ask_stop():
        return True

    algo.setStopCallback(ask_stop)
    algo.run()
    result = algo.getResult()
    status = algo.getResult().getStatus()
    msg = algo.getResult().getStatusMessage()
    calls = algo.getResult().getCallsNumber()
    print(f"{name}: {status} {msg} {calls}")
    assert result.getIterationNumber() <= 1



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

distribution = ot.Normal([0.5], [2.0], ot.CorrelationMatrix(1))
sample = distribution.getSample(200)

# Construct empirical CDF for the sample
data = ot.Sample(20, 2)
cursor = ot.Point(2)
for i in range(20):
    cursor[0] = 13.0 * i / 20 - 6.5
    count = 0.0
    for j in range(200):
        if sample[j, 0] < cursor[0]:
            count += 1.0
    cursor[1] = count / 200
    data[i] = cursor

# Create an empty graph
graph = ot.Graph("Staircase example", "x1", "x2")
graph.setLegendPosition("upper left")

# Create the staircase
staircase = ot.Staircase(data, "blue", "solid", "s", "Normal CDF")

# Then, draw it
graph.add(staircase)

fig = plt.figure(figsize=(4.5, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)



"""
Example of sensitivity analyses on the wing weight model
=========================================================
"""

# %%
#
# This example is a brief overview of the use of the most usual sensitivity analysis techniques and how to call them:
#
# - PCC: Partial Correlation Coefficients
# - PRCC: Partial Rank Correlation Coefficients
# - SRC: Standard Regression Coefficients
# - SRRC: Standard Rank Regression Coefficients
# - Pearson coefficients
# - Spearman coefficients
# - Taylor expansion importance factors
# - Sobol' indices
# - Rank-based estimation of Sobol' indices
# - HSIC : Hilbert-Schmidt Independence Criterion
#
# We present the methods on the :ref:`WingWeight function<use-case-wingweight>` and use the same notations.

# %%
# Definition of the model
# -----------------------
#
# We load the model from the usecases module.
#
#
import openturns as ot
import openturns.viewer as otv
from openturns.usecases.wingweight_function import WingWeightModel

m = WingWeightModel()

# %%
# Cross cuts of the function
# --------------------------
#
# Let's have a look on 2D cross cuts of the wing weight function.
# For each 2D cross cut, the other variables are fixed to the input distribution mean values.
# This graph allows one to have a first idea of the variations of the function in pair of dimensions.
# The colors of each contour plot are comparable.

lowerBound = m.distribution.getRange().getLowerBound()
upperBound = m.distribution.getRange().getUpperBound()

nX = ot.ResourceMap.GetAsUnsignedInteger("Evaluation-DefaultPointNumber")
description = m.distribution.getDescription()
description.add("")
m.model.setDescription(description)
m.model.setName("wing weight model")
grid = m.model.drawCrossCuts(
    m.distribution.getMean(),
    lowerBound,
    upperBound,
    [nX] * m.model.getInputDimension(),
    False,
    True,
    176.0,
    363.0,
)
grid.setTitle("")
# Get View object to manipulate the underlying figure
# Here we decide the colormap and the number of levels used for all contours
view = otv.View(grid, contour_kw={"cmap": "hsv", "levels": 55})

axes = view.getAxes()
fig = view.getFigure()
fig.set_size_inches(12, 12)  # reduce the size

# Setup a large colorbar
fig.colorbar(
    view.getSubviews()[1][0].getContourSets()[0], ax=axes[:-2, -1], fraction=0.3
)
# Hide unwanted axes labels
for i in range(len(axes)):
    for j in range(i + 1):
        if i < len(axes) - 1:
            axes[i][j].xaxis.set_ticklabels([])
        if j > 0:
            axes[i][j].yaxis.set_ticklabels([])
fig.subplots_adjust(top=0.99, bottom=0.05, left=0.06, right=0.99)

# %%
# We can see that the variables :math:`t_c, N_z, A, W_{dg}` seem to be influent on the wing weight whereas :math:`\Lambda, \ell, q, W_p, W_{fw}` have less influence on the function.

# %%
# Data generation
# ---------------
#
# We create the input and output data for the estimation of the different sensitivity coefficients and we get the input variables description:

inputNames = m.distribution.getDescription()

size = 500
inputDesign = m.distribution.getSample(size)
outputDesign = m.model(inputDesign)

# %%
# Let's estimate the PCC, PRCC, SRC, SRRC, Pearson and Spearman coefficients, display and analyze them.
# We create a :class:`~openturns.CorrelationAnalysis` model.

corr_analysis = ot.CorrelationAnalysis(inputDesign, outputDesign)

# %%
# PCC coefficients
# ----------------
# We compute here PCC coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
pcc_indices = corr_analysis.computePCC()
print(pcc_indices)

# %%
#

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    pcc_indices, inputNames, "PCC coefficients - Wing weight"
)
view = otv.View(graph)

# %%
# PRCC coefficients
# -----------------
# We compute here PRCC coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
prcc_indices = corr_analysis.computePRCC()
print(prcc_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    prcc_indices, inputNames, "PRCC coefficients - Wing weight"
)
view = otv.View(graph)

# %%
# SRC coefficients
# -------------------
# We compute here SRC coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
src_indices = corr_analysis.computeSRC()
print(src_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    src_indices, inputNames, "SRC coefficients - Wing weight"
)
view = otv.View(graph)

# %%
# Normalized squared SRC coefficients (coefficients are made to sum to 1) :

# %%
squared_src_indices = corr_analysis.computeSquaredSRC(True)
print(squared_src_indices)

# %%
# And their associated graph:

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    squared_src_indices, inputNames, "Squared SRC coefficients - Wing weight"
)
view = otv.View(graph)

# %%
#

# %%
# SRRC coefficients
# --------------------
# We compute here SRRC coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
srrc_indices = corr_analysis.computeSRRC()
print(srrc_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    srrc_indices, inputNames, "SRRC coefficients - Wing weight"
)
view = otv.View(graph)

# %%
# Pearson coefficients
# -----------------------
# We compute here the Pearson :math:`\rho` coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
pearson_correlation = corr_analysis.computeLinearCorrelation()
print(pearson_correlation)

# %%
title_pearson_graph = "Pearson correlation coefficients - Wing weight"
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    pearson_correlation, inputNames, title_pearson_graph
)
view = otv.View(graph)

# %%
# Spearman coefficients
# -----------------------
# We compute here the Spearman :math:`\rho_s` coefficients using the :class:`~openturns.CorrelationAnalysis`.

# %%
spearman_correlation = corr_analysis.computeSpearmanCorrelation()
print(spearman_correlation)

# %%
title_spearman_graph = "Spearman correlation coefficients - Wing weight"
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    spearman_correlation, inputNames, title_spearman_graph
)
view = otv.View(graph)

# %%
#
# The different computed correlation estimators show that the variables :math:`S_w, A, N_z, t_c` seem to be the most correlated with the wing weight in absolute value.
# Pearson and Spearman coefficients do not reveal any linear nor monotonic correlation as no coefficients are equal to +/- 1.
# Coefficients about :math:`t_c` are negative revealing a negative correlation with the wing weight, that is consistent with the model expression.


# %%
# Taylor expansion importance factors
# -----------------------------------
# We compute here the Taylor expansion importance factors using :class:`~openturns.TaylorExpansionMoments`.

# %%

# %%
# We create a distribution-based RandomVector.
X = ot.RandomVector(m.distribution)

# %%
# We create a composite RandomVector Y from X and m.model.
Y = ot.CompositeRandomVector(m.model, X)

# %%
# We create a Taylor expansion method to approximate moments.
taylor = ot.TaylorExpansionMoments(Y)

# %%
# We get the importance factors.
print(taylor.getImportanceFactors())

# %%
# We draw the importance factors
graph = taylor.drawImportanceFactors()
graph.setTitle("Taylor expansion imporfance factors - Wing weight")
view = otv.View(graph)

# %%
#
# The Taylor expansion importance factors is consistent with the previous estimators as :math:`S_w, A, N_z, t_c` seem to be the most influent variables.
# To analyze the relevance of the previous indices, a Sobol' analysis is now carried out.

# %%
# Sobol' indices
# --------------
# We compute the Sobol' indices from both sampling approach and  Polynomial Chaos Expansion.

# %%
sizeSobol = 1000
sie = ot.SobolIndicesExperiment(m.distribution, sizeSobol)
inputDesignSobol = sie.generate()
inputNames = m.distribution.getDescription()
inputDesignSobol.setDescription(inputNames)
inputDesignSobol.getSize()

# %%
# We see that 12000 function evaluations are required to estimate the first order and total Sobol' indices.

# %%
# Then, we evaluate the outputs corresponding to this design of experiments.

# %%
outputDesignSobol = m.model(inputDesignSobol)

# %%
# We estimate the Sobol' indices with the :class:`~openturns.SaltelliSensitivityAlgorithm`.

# %%
sensitivityAnalysis = ot.SaltelliSensitivityAlgorithm(
    inputDesignSobol, outputDesignSobol, sizeSobol
)

# %%
# The `getFirstOrderIndices` and `getTotalOrderIndices` methods respectively return estimates of all first order and total Sobol' indices.

# %%
print("First order indices:", sensitivityAnalysis.getFirstOrderIndices())

# %%
print("Total order indices:", sensitivityAnalysis.getTotalOrderIndices())


# %%
# The `draw` method produces the following graph. The vertical bars represent the 95% confidence intervals of the estimates.

# %%
graph = sensitivityAnalysis.draw()
graph.setTitle("Sobol indices with Saltelli - wing weight")
view = otv.View(graph)

# %%
# We see that several Sobol' indices are negative, that is inconsistent with the theory. Therefore, a larger number of samples is required to get consistent indices
sizeSobol = 10000
sie = ot.SobolIndicesExperiment(m.distribution, sizeSobol)
inputDesignSobol = sie.generate()
inputNames = m.distribution.getDescription()
inputDesignSobol.setDescription(inputNames)
inputDesignSobol.getSize()
outputDesignSobol = m.model(inputDesignSobol)

sensitivityAnalysis = ot.SaltelliSensitivityAlgorithm(
    inputDesignSobol, outputDesignSobol, sizeSobol
)

sensitivityAnalysis.getFirstOrderIndices()
sensitivityAnalysis.getTotalOrderIndices()

graph = sensitivityAnalysis.draw()
graph.setTitle("Sobol indices with Saltelli - wing weight")
view = otv.View(graph)

# %%
# It improves the accuracy of the estimation but, for very low indices, Saltelli scheme is not accurate since several confidence intervals provide negative lower bounds.

# %%
# Now, we estimate the Sobol' indices using Polynomial Chaos Expansion.
# We create a Functional Chaos Expansion.
sizePCE = 800
inputDesignPCE = m.distribution.getSample(sizePCE)
outputDesignPCE = m.model(inputDesignPCE)

algo = ot.FunctionalChaosAlgorithm(inputDesignPCE, outputDesignPCE, m.distribution)

algo.run()
result = algo.getResult()

# %%
# Then, we exploit the surrogate model to compute the Sobol' indices.
sensitivityAnalysis = ot.FunctionalChaosSobolIndices(result)
sensitivityAnalysis

# %%
firstOrder = [sensitivityAnalysis.getSobolIndex(i) for i in range(m.dim)]
totalOrder = [sensitivityAnalysis.getSobolTotalIndex(i) for i in range(m.dim)]
graph = ot.SobolIndicesAlgorithm.DrawSobolIndices(inputNames, firstOrder, totalOrder)
graph.setTitle("Sobol indices by Polynomial Chaos Expansion - wing weight")
view = otv.View(graph)


# %%
# Furthermore, first order Sobol' indices can also been estimated in a data-driven way using a rank-based sensitivity algorithm.
# In such a way, the estimation of sensitivity indices does not involve any surrogate model.
sizeRankSobol = 800
inputDesignRankSobol = m.distribution.getSample(sizeRankSobol)
outputDesignankSobol = m.model(inputDesignRankSobol)
myRankSobol = ot.RankSobolSensitivityAlgorithm(
    inputDesignRankSobol, outputDesignankSobol
)
indicesrankSobol = myRankSobol.getFirstOrderIndices()
print("First order indices:", indicesrankSobol)
graph = myRankSobol.draw()
graph.setTitle("Sobol indices by rank-based estimation - wing weight")
view = otv.View(graph)

# %%
#
# The Sobol' indices confirm the previous analyses, in terms of ranking of the most influent variables.
# We also see that five variables have a quasi null total Sobol' indices, that indicates almost no influence on the wing weight.
# There is no discrepancy between first order and total Sobol' indices, that indicates no or very low interaction between the variables in the variance of the output.
# As the most important variables act only through decoupled first degree contributions, the hypothesis of a linear dependence between the input variables and the weight is legitimate.
# This explains why both squared SRC and Taylor give the exact same results even if the first one is based on a :math:`\mathcal{L}^2` linear approximation
# and the second one is based on a linear expansion around the mean value of the input variables.


# %%
# HSIC indices
# ------------

# %%
# We then estimate the HSIC indices using a data-driven approach.
sizeHSIC = 250
inputDesignHSIC = m.distribution.getSample(sizeHSIC)
outputDesignHSIC = m.model(inputDesignHSIC)

covarianceModelCollection = []

# %%
for i in range(m.dim):
    Xi = inputDesignHSIC.getMarginal(i)
    inputCovariance = ot.SquaredExponential(1)
    inputCovariance.setScale(Xi.computeStandardDeviation())
    covarianceModelCollection.append(inputCovariance)

# %%
# We define a covariance kernel associated to the output variable.
outputCovariance = ot.SquaredExponential(1)
outputCovariance.setScale(outputDesignHSIC.computeStandardDeviation())
covarianceModelCollection.append(outputCovariance)

# %%
# In this paragraph, we perform the analysis on the raw data: that is
# the global HSIC estimator.
estimatorType = ot.HSICUStat()

# %%
# We now build the HSIC estimator:
globHSIC = ot.HSICEstimatorGlobalSensitivity(
    covarianceModelCollection, inputDesignHSIC, outputDesignHSIC, estimatorType
)

# %%
# We get the R2-HSIC indices:
R2HSICIndices = globHSIC.getR2HSICIndices()
print("\n Global HSIC analysis")
print("R2-HSIC Indices: ", R2HSICIndices)

# %%
# and the HSIC indices:
HSICIndices = globHSIC.getHSICIndices()
print("HSIC Indices: ", HSICIndices)

# %%
# The p-value by permutation.
pvperm = globHSIC.getPValuesPermutation()
print("p-value (permutation): ", pvperm)

# %%
# We have an asymptotic estimate of the value for this estimator.
pvas = globHSIC.getPValuesAsymptotic()
print("p-value (asymptotic): ", pvas)

# %%
# We vizualise the results.
graph1 = globHSIC.drawHSICIndices()
view1 = otv.View(graph1)

graph2 = globHSIC.drawPValuesAsymptotic()
view2 = otv.View(graph2)

graph3 = globHSIC.drawR2HSICIndices()
view3 = otv.View(graph3)

graph4 = globHSIC.drawPValuesPermutation()
view4 = otv.View(graph4)

# %%
# The HSIC indices go in the same way as the other estimators in terms the most influent variables.
# The variables :math:`W_{fw}, q, l, W_p` seem to be independent to the output as the corresponding p-values are high.
# We can also see that the asymptotic p-values and p-values estimated by permutation are quite similar.

# %%
otv.View.ShowAll()



"""
Optimize an LHS design of experiments
=====================================
"""

# %%
# This examples show how to generate optimized LHS experiments according to the different criteria.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# **LHS and space filling**

# %%
N = 100
# Considering independent Uniform distributions of dimension 3
# Bounds are (-1,1), (0,2) and (0, 0.5)
distribution = ot.JointDistribution(
    [ot.Uniform(-1.0, 1.0), ot.Uniform(0.0, 2.0), ot.Uniform(0.0, 0.5)]
)
# Random LHS
lhs = ot.LHSExperiment(distribution, N)
lhs.setAlwaysShuffle(True)  # randomized
design = lhs.generate()
# C2
c2 = ot.SpaceFillingC2().evaluate(design)
# PhiP with default p
phip = ot.SpaceFillingPhiP().evaluate(design)
# mindist
mindist = ot.SpaceFillingMinDist().evaluate(design)
# For p->infinity
phip_inf = ot.SpaceFillingPhiP(100).evaluate(design)
print(phip, mindist, phip_inf)

# %%
# **Optimized LHS using Monte Carlo**

# %%
# As with Monte Carlo, user decides of a fixed number of iterations,
# but this time this number is part of the temperature profile.
#
# Two profiles are currently provided:
# - Linear profile: :math:`T_i = T_0 (1-\frac{1}{n_{iter}})`
# - Geometric profile: :math:`T_i = T_O c^i, 0<c<1`
#
# Starting from an LHS design, a new design is built by permuting a random coordinate of two randomly chosen sample points;
# this new design is also an LHS, but not necessarily a more efficient design.
#
# A comparison of criteria of the two designs is done, and the new LHS is accepted with probability
#
# .. math::
#    min\left(exp{\left[-\frac{\Phi(LHS_{new}) - \Phi(LHS)}{T_i}\right]}, 1\right)
#

# %%
# Considering independent Uniform(0,1) distributions of dimension 3
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
# Random LHS
lhs = ot.LHSExperiment(distribution, N)
lhs.setAlwaysShuffle(True)  # randomized
algo = ot.SimulatedAnnealingLHS(lhs)
design = algo.generate()

# %%
# One could also fix the criterion, the temperature profile and get more results.

# Considering independent Uniform distributions of dimension 3
# Bounds are (-1,1), (0,2) and (0, 0.5)
distribution = ot.JointDistribution(
    [ot.Uniform(-1.0, 1.0), ot.Uniform(0.0, 2.0), ot.Uniform(0.0, 0.5)]
)
# Random LHS
lhs = ot.LHSExperiment(distribution, N)
lhs.setAlwaysShuffle(True)  # randomized
# Fixing C2 crit
space_filling = ot.SpaceFillingC2()
# Defining a temperature profile
# A geometric profile seems accurate with default parameters
# e.g. T0=10, c=0.95, iMax=2000
temperatureProfile = ot.GeometricProfile()
algo = ot.SimulatedAnnealingLHS(lhs, space_filling, temperatureProfile)
# optimal design
design = algo.generate()
result = algo.getResult()
# Criteria for the optimal design
crit_c2 = result.getC2()
crit_phip = result.getPhiP()
crit_mindist = result.getMinDist()
# History of the criterion used for optimization
history = result.getAlgoHistory()
criterion_hist = history[:, 0]
# Additional results
temperature_hist = history[:, 1]
probability_hist = history[:, 2]

# %%
# It is also possible to chain several iterations of the whole process with different starting points.
N = 10

# Considering independent Uniform distributions of dimension 3
# Bounds are (-1,1), (0,2) and (0, 0.5)
distribution = ot.JointDistribution(
    [ot.Uniform(-1.0, 1.0), ot.Uniform(0.0, 2.0), ot.Uniform(0.0, 0.5)]
)
# Random LHS
lhs = ot.LHSExperiment(distribution, N)
lhs.setAlwaysShuffle(True)  # randomized
# Fixing PhiP crit
space_filling = ot.SpaceFillingPhiP()
# Defining a temperature profile
# T0=10, iMax=3000
temperatureProfile = ot.LinearProfile(10.0, 3000)
algo = ot.SimulatedAnnealingLHS(lhs, space_filling, temperatureProfile)
restart = 50
design = algo.generateWithRestart(restart)
# Retrieve all optimal designs
result = algo.getResult()
designs = [result.getOptimalDesign(i) for i in range(restart)]

# %%
# Finally, we could start the optimization process of LHS using a precomputed LHS design.

# Considering independent Uniform distributions of dimension 3
# Bounds are (0,1)^3
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
# Random LHS
lhs = ot.LHSExperiment(distribution, N)
lhs.setAlwaysShuffle(True)  # randomized
# Fixing C2 crit for example
space_filling = ot.SpaceFillingC2()
# Defining a temperature profile
# T0=10, iMax=3000
temperatureProfile = ot.LinearProfile(10.0, 3000)
algo = ot.SimulatedAnnealingLHS(lhs, space_filling, temperatureProfile)
design = algo.generate()
result = algo.getResult()
# check history ==> draw criterion
graph = result.drawHistoryCriterion()
view = otv.View(graph)

# %%
# Convergence needs to be performed
# New algo starting from this design
algo = ot.SimulatedAnnealingLHS(design, distribution, space_filling, temperatureProfile)
design = algo.generate()

# %%
# Display all figures
otv.View.ShowAll()



"""
Create a polynomial chaos metamodel from a data set
===================================================
"""

# %%
# In this example, we create a polynomial chaos expansion (PCE) using
# a data set.
# More precisely, given a pair of input and output samples,
# we create a PCE without the knowledge of the input distribution.
# In this example, we use a relatively small sample size equal to 80.

# %%
# In this example we create a global approximation of a model response using
# polynomial chaos expansion.
#
# Let :math:`\vect{g}` be the function defined by:
#
# .. math::
#    \vect{g}(\vect{x}) = \Tr{\left(\cos(x_1 + x_2), (x_2 + 1) e^{x_1}\right)}
#
#
# for any :math:`\vect{x} \in \Rset^2`.
#
# We assume that
#
# .. math::
#    X_1 \sim \mathcal{N}(0,1) \textrm{ and } X_2 \sim \mathcal{N}(0,1)
#
# and that :math:`X_1` and :math:`X_2` are independent.
#
# An interesting point in this example is that the output is multivariate.
# This is why we are going to use the `getMarginal` method in the script
# in order to select the output marginal that we want to manage.

# %%
# Simulate input and output samples
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# %%
import openturns as ot
import openturns.viewer as otv
from matplotlib import pyplot as plt


# %%
# We first create the function `model`.

# %%
input_names = ["x1", "x2"]
formulas = ["cos(x1 + x2)", "(x2 + 1) * exp(x1)"]
model = ot.SymbolicFunction(input_names, formulas)
inputDimension = model.getInputDimension()
outputDimension = model.getOutputDimension()

# %%
# Then we create a sample `inputSample` and compute the corresponding output
# sample `outputSample`.

# %%
distribution = ot.Normal(inputDimension)
samplesize = 80
inputSample = distribution.getSample(samplesize)
outputSample = model(inputSample)

# %%
# Create the PCE
# ~~~~~~~~~~~~~~

# %%
# Create a functional chaos model.
# The algorithm needs to fit a distribution on the input sample.
# To do this, the algorithm in :class:`~openturns.FunctionalChaosAlgorithm`
# uses the :class:`~openturns.FunctionalChaosAlgorithm.BuildDistribution`
# static method to fit the distribution to the input sample.
# Please read :doc:`Fit a distribution from an input sample </auto_surrogate_modeling/polynomial_chaos/plot_chaos_build_distribution>`
# for an example of this method.
# The algorithm does this automatically using the Lilliefors test.
# In order to make the algorithm a little faster, we reduce the
# value of the sample size used in the Lilliefors test.

# %%
ot.ResourceMap.SetAsUnsignedInteger("FittingTest-LillieforsMaximumSamplingSize", 50)

# %%
# The main topic of this example is to introduce the next constructor of
# :class:`~openturns.FunctionalChaosAlgorithm`.
# Notice that the only input arguments are the input and output samples.
algo = ot.FunctionalChaosAlgorithm(inputSample, outputSample)
algo.run()
result = algo.getResult()
result

# %%
# Not all coefficients are selected in this PCE.
# Indeed, the default constructor of :class:`~openturns.FunctionalChaosAlgorithm`
# creates a sparse PCE.
# Please read :doc:`Create a full or sparse polynomial chaos expansion </auto_surrogate_modeling/polynomial_chaos/plot_functional_chaos_database>`
# for more details on this topic.

# %%
# Get the metamodel.
metamodel = result.getMetaModel()

# %%
# Plot the second output of our model depending on :math:`x_2` with :math:`x_1=0.5`.
# In order to do this, we create a `ParametricFunction` and set the value of :math:`x_1`.
# Then we use the `getMarginal` method to extract the second output (which index is equal to 1).

# %%
x1index = 0
x1value = 0.5
x2min = -3.0
x2max = 3.0
outputIndex = 1
metamodelParametric = ot.ParametricFunction(metamodel, [x1index], [x1value])
graph = metamodelParametric.getMarginal(outputIndex).draw(x2min, x2max)
graph.setLegends(["Metamodel"])
modelParametric = ot.ParametricFunction(model, [x1index], [x1value])
curve = modelParametric.getMarginal(outputIndex).draw(x2min, x2max).getDrawable(0)
curve.setColor("red")
curve.setLegend("Model")
graph.add(curve)
graph.setLegendPosition("lower right")
graph.setXTitle("X2")
graph.setTitle("Metamodel Validation, output #%d" % (outputIndex))
view = otv.View(graph)

# %%
# We see that the metamodel fits approximately to the model, except
# perhaps for extreme values of :math:`x_2`.
# However, there is a better way of globally validating the metamodel,
# using the :class:`~openturns.MetaModelValidation` on a validation design of experiments.

# %%
n_valid = 100
inputTest = distribution.getSample(n_valid)
outputTest = model(inputTest)


# %%
# Plot the corresponding validation graphics.

# %%
metamodelPredictions = metamodel(inputTest)
val = ot.MetaModelValidation(outputTest, metamodelPredictions)
r2Score = val.computeR2Score()
graph = val.drawValidation()
graph.setTitle("Metamodel validation R2=" + str(r2Score))
view = otv.View(graph)

# %%
# The coefficient of determination is not extremely satisfactory for the
# first output, but is would be sufficient for a central dispersion study.
# The second output has a much more satisfactory :math:`R^2`: only one single
# extreme point is far from the diagonal of the graphics.

# %%
# Compute and print Sobol' indices
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# %%
chaosSI = ot.FunctionalChaosSobolIndices(result)
chaosSI

# %%
# Let us analyze the results of this global sensitivity analysis.
#
# * We see that the first output involves significant multi-indices with
#   higher marginal degree.
# * For the second output, the first variable is especially significant,
#   with a significant contribution of the interactions.
#   The contribution of the interactions are very
#   significant in this model.

# %%
# Draw Sobol' indices.

# %%
sensitivityAnalysis = ot.FunctionalChaosSobolIndices(result)
first_order = [sensitivityAnalysis.getSobolIndex(i) for i in range(inputDimension)]
total_order = [sensitivityAnalysis.getSobolTotalIndex(i) for i in range(inputDimension)]

# %%
input_names = model.getInputDescription()
graph = ot.SobolIndicesAlgorithm.DrawSobolIndices(input_names, first_order, total_order)
graph.setLegendPosition("center")
view = otv.View(graph)

# %%
# Testing the sensitivity to the degree
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# With the specific constructor of :class:`~openturns.FunctionalChaosAlgorithm` that
# we use, the `FunctionalChaosAlgorithm-MaximumTotalDegree`
# in the `ResourceMap` configures the maximum degree explored by
# the algorithm. This degree is a trade-off.
#
# * If the maximum degree is too low, the polynomial may miss some
#   coefficients so that the quality is lower than possible.
# * If the maximum degree is too large, the number of coefficients
#   to explore is too large, so that the coefficients might be poorly estimated.
#
# This is why the following `for` loop explores various degrees to see
# the sensitivity of the metamodel predictivity depending on the degree.

# %%
# The default value of this parameter is 10.

# %%
ot.ResourceMap.GetAsUnsignedInteger("FunctionalChaosAlgorithm-MaximumTotalDegree")

# %%
# This is why we explore the values from 1 to 10.

# %%
maximumDegree = 11
degrees = range(1, maximumDegree)
r2Score = ot.Sample(len(degrees), outputDimension)
for maximumDegree in degrees:
    ot.ResourceMap.SetAsUnsignedInteger(
        "FunctionalChaosAlgorithm-MaximumTotalDegree", maximumDegree
    )
    print("Maximum total degree =", maximumDegree)
    algo = ot.FunctionalChaosAlgorithm(inputSample, outputSample)
    algo.run()
    result = algo.getResult()
    metamodel = result.getMetaModel()
    metamodelPredictions = metamodel(inputTest)
    val = ot.MetaModelValidation(outputTest, metamodelPredictions)
    r2ScoreLocal = val.computeR2Score()
    r2ScoreLocal = [max(0.0, r2ScoreLocal[i]) for i in range(outputDimension)]
    r2Score[maximumDegree - degrees[0]] = r2ScoreLocal

# %%
graph = ot.Graph("Predictivity", "Total degree", "R2")
cloud = ot.Cloud([[d] for d in degrees], r2Score[:, 0])
cloud.setLegend("Output #0")
cloud.setPointStyle("bullet")
graph.add(cloud)
cloud = ot.Cloud([[d] for d in degrees], r2Score[:, 1])
cloud.setLegend("Output #1")
cloud.setPointStyle("diamond")
graph.add(cloud)
graph.setLegendPosition("upper left")
graph.setLegendCorner([1.0, 1.0])
view = otv.View(graph)
plt.subplots_adjust(right=0.7)

# %%
# We see that a low total degree is not sufficient to describe the
# first output with good :math:`R^2` score.
# However, the coefficient of determination can drop when the total degree increases.
# The :math:`R^2` score of the second output seems to be much less satisfactory:
# a little more work would be required to improve the metamodel.
#
# In this situation, the following methods may be used.
#
# * Since the distribution of the input is known, we may want to give
#   this information to the :class:`~openturns.FunctionalChaosAlgorithm`.
#   This prevents the algorithm from trying to fit the input distribution
#   which best fit to the data.
# * We may want to customize the `adaptiveStrategy` by selecting an enumerate
#   function which best fit to this particular example.
#   In this specific example, however, the interactions plays a great role so that the
#   linear enumerate function may provide better results than the hyperbolic rule.
# * We may want to customize the `projectionStrategy` by selecting a method
#   to compute the coefficient which improves the estimation.
#   For example, it might be interesting to
#   try an integration rule instead of the least squares method.
#   Notice that a specific design of experiments is required in this case.

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
from math import sin

ot.TESTPREAMBLE()


# lm build
print("Fit y ~ 3 - 2 x + 0.05 * sin(x) model using 20 points (sin(x) ~ noise)")
size = 20
oneSample = ot.Sample(size, 1)
twoSample = ot.Sample(size, 1)
for i in range(size):
    oneSample[i, 0] = 7.0 * sin(-3.5 + (6.5 * i) / (size - 1.0)) + 2.0
    twoSample[i, 0] = -2.0 * oneSample[i, 0] + 3.0 + 0.05 * sin(oneSample[i, 0])

test = ot.LinearModelAlgorithm(oneSample, twoSample)
result = ot.LinearModelResult(test.getResult())
analysis = ot.LinearModelAnalysis(result)
print(analysis)
# Compute confidence level (95%) for coefficients estimate
alpha = 0.95
# interval confidence bounds
interval = analysis.getCoefficientsConfidenceInterval(alpha)
print("confidence intervals with level=%1.2f : %s" % (alpha, interval))

print("")
print("")
print("")
print("Fit y ~ 1 + 0.1 x + 10 x^2 model using 100 points")
ot.RandomGenerator.SetSeed(0)
size = 100
# Define a linespace from 0 to 10 with size points
# We use a Box experiment ==> remove 0 & 1 points
experiment = ot.Box([size - 2])
X = experiment.generate()
# X is defined in [0,1]
X *= [10]
# Stack X2
X2 = ot.Sample(X)
for i in range(size):
    X2[i, 0] = X[i, 0] * X2[i, 0]
X.stack(X2)
# Define y
Y = ot.Sample(size, 1)
for i in range(size):
    Y[i, 0] = (
        1.0 + 0.1 * X[i, 0] + 10.0 * X[i, 0] * X[i, 0] + 0.1 * ot.DistFunc.rNormal()
    )
test = ot.LinearModelAlgorithm(X, Y)
result = test.getResult()
analysis = ot.LinearModelAnalysis(result)
print(analysis)
# Compute confidence level (95%) for coefficients estimate
alpha = 0.95
# interval confidence bounds
interval = analysis.getCoefficientsConfidenceInterval(alpha)
print("confidence intervals with level=%1.2f : %s" % (alpha, interval))

# https://github.com/openturns/openturns/issues/1729
ot.RandomGenerator.SetSeed(1789)
sample_size = 14605  # 14604 : OK, 14605 : Fail
a0 = 0.6
a1 = 2.0
x = ot.Normal(0.5, 0.2).getSample(sample_size)
epsilon = ot.Exponential(0.1).getSample(sample_size)
y = ot.Sample([[a0]] * sample_size) + a1 * x + epsilon
algo = ot.LinearModelAlgorithm(x, y)
result = algo.getResult()
analysis = ot.LinearModelAnalysis(result)
results = str(analysis)

# Fix #1820
ot.RandomGenerator.SetSeed(0)
distribution = ot.Normal()
sample = distribution.getSample(30)
func = ot.SymbolicFunction("x", "2 * x + 1")
firstSample = sample
secondSample = func(sample) + ot.Normal().getSample(30)
linear_model = ot.LinearModelAlgorithm(firstSample, secondSample)
linear_result = linear_model.getResult()
test_result = ot.LinearModelTest.LinearModelFisher(
    firstSample, secondSample, linear_result
)
linear_model_analysis = ot.LinearModelAnalysis(linear_result)
ott.assert_almost_equal(
    test_result.getStatistic(), linear_model_analysis.getFisherScore(), 1e-13, 0
)
ott.assert_almost_equal(
    test_result.getPValue(), linear_model_analysis.getFisherPValue(), 0, 0
)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Construction based on two functions
# Create an instance
grid = ot.RegularGrid(0.0, 0.1, 11)
myFunc = ot.FieldToFieldConnection(
    ot.ValueFunction(ot.SymbolicFunction("x", "sin(x)"), grid),
    ot.ValueFunction(ot.SymbolicFunction("x", "cos(x)"), grid),
)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point
field = ot.Field(grid, grid.getVertices() * [2.0])
print("field=", field)
print("myFunc(field)=", myFunc(field))
# Get the number of calls
print("called ", myFunc.getCallsNumber(), " times")

# Construction based on a FieldToPointFunction followed by a PointToFieldFunction
# Create a KarhunenLoeveResult
mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 0.0)
algo.run()
result = algo.getResult()
# Create a PointToFieldFunction
lifting = ot.KarhunenLoeveLifting(result)
# Create a FieldToPointFunction
projection = ot.KarhunenLoeveProjection(result)
# Create an instance
myFunc = ot.FieldToFieldConnection(lifting, projection)

print("myFunc=", myFunc)
# Get the input and output description
print("myFunc input description=", myFunc.getInputDescription())
print("myFunc output description=", myFunc.getOutputDescription())
# Get the input and output dimension
print("myFunc input dimension=", myFunc.getInputDimension())
print("myFunc output dimension=", myFunc.getOutputDimension())
# Connection on a point
field = result.getModesAsProcessSample().computeMean()
print("field=", field)
print("myFunc(field)=", myFunc(field))
# Get the number of calls
print("called ", myFunc.getCallsNumber(), " times")



"""
Use the post-analytical importance sampling algorithm
=====================================================
"""

# %%
# In this example we want to estimate the probability to exceed a threshold through the combination of approximation and simulation methods.
#
#   - perform an FORM or SORM study in order to find the design point,
#   - perform an importance sampling study centered around the design point:
#     the importance distribution operates in the standard space and is the
#     standard distribution of the standard space (the standard elliptical
#     distribution in the case of an elliptic copula of the input random vector,
#     the standard normal one in all the other cases).
#
# The importance sampling technique in the standard space may be of two kinds:
#
#   - the sample is generated according to the new importance distribution:
#     this technique is called post analytical importance sampling,
#   - the sample is generated according to the new importance distribution and
#     is controlled by the value of the linearized limit state function:
#     this technique is called post analytical controlled importance sampling.
#
# This post analytical importance sampling algorithm is created from the result structure of a FORM or SORM algorithm.
#
# It is parameterized like other simulation algorithm, through the parameters OuterSampling, BlockSize, ... and provide the same type of results.
#
# Let us note that the post FORM/SORM importance sampling method may be
# implemented thanks to the ImportanceSampling object, where the importance
# distribution is defined in the standard space: then, it requires that the
# event initially defined in the pysical space be transformed in the standard space.
#
# The controlled importance sampling technique is only accessible within the post analytical context.

# %%
import openturns as ot


# %%
# Create a model
model = ot.SymbolicFunction(["x1", "x2"], ["x1^2+x2"])
R = ot.CorrelationMatrix(2)
R[0, 1] = -0.6
inputDist = ot.Normal([0.0, 0.0], R)
inputDist.setDescription(["X1", "X2"])
inputVector = ot.RandomVector(inputDist)

# Create the output random vector Y=model(X)
Y = ot.CompositeRandomVector(model, inputVector)

# Create the event Y > 4
threshold = 4.0
event = ot.ThresholdEvent(Y, ot.Greater(), threshold)

# %%
# Create a FORM algorithm
solver = ot.Cobyla()
solver.setStartingPoint(inputDist.getMean())
algo = ot.FORM(solver, event)

# Run the algorithm and retrieve the result
algo.run()
result_form = algo.getResult()

# %%
# Create the post analytical importance sampling simulation algorithm
algo = ot.PostAnalyticalImportanceSampling(result_form)
algo.run()
algo.getResult()

# %%
# Create the post analytical controlled importance sampling simulation algorithm
algo = ot.PostAnalyticalControlledImportanceSampling(result_form)
algo.run()
algo.getResult()



#! /usr/bin/env python

from matplotlib import pyplot as plt

from openturns.viewer import PlotDesign
import openturns as ot

samplesize = 12

# Plot 2D - basic
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design)

# Plot 3D - basic
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design)

# Plot 2D - set bounds and subdivisions
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 10])

# Plot 3D - set bounds and subdivisions
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 10, 10])

# Plot 2D - advanced - enable ticks
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 10], enableTicks=True)
fig.set_size_inches(12, 12)

# Plot 3D - advanced - enable ticks
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 10, 10], enableTicks=True)
fig.set_size_inches(12, 12)

# Plot 2D - advanced - unequal number of divisions
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 5])

# Plot 3D - advanced - unequal number of divisions
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 3)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = PlotDesign(design, bounds, [10, 5, 3])

# Plot 2D - set figure prior to calling the plot
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
bounds = distribution.getRange()
design = distribution.getSample(samplesize)
fig = plt.figure()
fig = PlotDesign(design, figure=fig)



"""
LOLA-Voronoi sequential design of experiment
============================================
"""

# %%
# The LOLA-Voronoi sequential experiment helps to generate an optimized design allowing
# better approximations of functions taking into account empty regions and gradient values.
# It can be relevant to build a design of experiment for a metamodel: we will compare the
# LOLA-Voronoi design against a Sobol' design as learning points for a chaos metamodel.

# %%
import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv

# %%
# Lets use Franke's bivariate function
dim = 2
f1 = ot.SymbolicFunction(
    ["a0", "a1"],
    [
        "3 / 4 * exp(-1 / 4 * (((9 * a0 - 2) ^ 2) + ((9 * a1 - 2) ^ 2))) + 3 / 4 * exp(-1 / 49 * "
        "((9 * a0 + 1) ^ 2) - 1 / 10 * (9 * a1 + 1) ^ 2) + 1 / 2 * exp(-1 / 4 * (((9 * a0 - 7) ^ 2) "
        "+ (9 * a1 - 3) ^ 2)) - 1 / 5 * exp(-((9 * a0 - 4) ^ 2) - ((9 * a1 + 1) ^ 2))"
    ],
)
print(f1([0.5, 0.5]))
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)

# %%
# Plot the function
ot.ResourceMap.SetAsString("Contour-DefaultColorMapNorm", "rank")
graph = f1.draw(
    distribution.getRange().getLowerBound(), distribution.getRange().getUpperBound()
)
contour = graph.getDrawable(0)
contour.setLegend("model")
graph.setTitle("Model")
graph.setXTitle("x1")
graph.setYTitle("x2")
_ = otv.View(graph, square_axes=True)

# %%
# Plot the hessian norm


def pyHessianNorm(X):
    h = f1.hessian(X).getSheet(0)
    h.squareElements()
    s = h.computeSumElements()
    return [s**0.5]


hessNorm = ot.PythonFunction(f1.getInputDimension(), 1, pyHessianNorm)
graph = hessNorm.draw(
    distribution.getRange().getLowerBound(), distribution.getRange().getUpperBound()
)
graph.setTitle("Hessian norm")
graph.setXTitle("x1")
graph.setYTitle("x2")
_ = otv.View(graph, square_axes=True)

# %%
# Lets define an initial design of experiments
N = 50
x0 = ot.LowDiscrepancyExperiment(ot.HaltonSequence(), distribution, N).generate()
y0 = f1(x0)

# %%
# Plot the initial input sample
graph = ot.Graph(f"Initial points N={N}", "x1", "x2")
initial = ot.Cloud(x0)
initial.setPointStyle("fcircle")
initial.setColor("blue")
initial.setLegend(f"initial ({len(x0)})")
graph.add(initial)
graph.add(contour)
_ = otv.View(graph, square_axes=True)

# %%
# Instantiate the algorithm from the initial DOE and the distribution
# The algorithm can be fined-tuned using keys in ResourceMap, see
# the API documentation. It can have a major impact on the performance
# of the algorithm. In particular, if your distribution has independent
# components you may set "LOLAVoronoi-UseTruncatedDistribution" to True.
algo = otexp.LOLAVoronoi(x0, y0, distribution)

# %%
# Iteratively generate new samples: add 50 points, in 10 blocks of 5 points.
inc = 5
contour = contour.getImplementation()
contour.setColorBarPosition("")  # hide color bar
for i in range(10):
    graph = ot.Graph("", "x1", "x2")
    graph.setLegendPosition("upper left")
    graph.setLegendFontSize(8)
    graph.add(contour)
    graph.add(initial)

    if i > 0:
        previous = ot.Cloud(algo.getInputSample()[len(x0) : N])
        previous.setPointStyle("fcircle")
        previous.setColor("red")
        previous.setLegend(f"previous iterations ({N - len(x0)})")
        graph.add(previous)

    x = algo.generate(inc)
    y = f1(x)
    algo.update(x, y)
    N = algo.getGenerationIndices()[-1]

    current = ot.Cloud(x)
    current.setPointStyle("fcircle")
    current.setColor("orange")
    current.setLegend(f"current iteration ({inc})")
    graph.add(current)
    graph.setTitle(f"LOLA-Voronoi iteration #{i + 1} N={N}")
    otv.View(graph, square_axes=True)


# %%
# Lets compare metamodels from LOLA samples versus other design
xLola, yLola = algo.getInputSample(), algo.getOutputSample()
learnSize = xLola.getSize()


def runMetaModel(x, y, tag):
    algo = ot.LeastSquaresExpansion(x, y, distribution)
    algo.run()
    metamodel = algo.getResult().getMetaModel()
    yPred = metamodel(xRef)
    validation = ot.MetaModelValidation(yRef, yPred)
    mse = validation.computeMeanSquaredError()
    maxerr = (yRef - yPred).asPoint().normInf()
    print(f"{tag} mse={mse} r2={validation.computeR2Score()} maxerr={maxerr:.3f}")


# %%
# Generate a large validation sample by Monte Carlo
nRef = int(1e6)
xRef = distribution.getSample(nRef)
yRef = f1(xRef)

# %%
# Build a metamodel from Sobol' samples
xSobol = ot.LowDiscrepancyExperiment(
    ot.SobolSequence(), distribution, learnSize
).generate()
ySobol = f1(xSobol)
runMetaModel(xSobol, ySobol, "Sobol")

# %%
# Build a metamodel on the LOLA global samples
# We observe that the metamodel error metrics (MSE, R2) and maximum error
# from the LOLA design are a bit better compared to the Sobol experiment
runMetaModel(xLola, yLola, "LOLA")

# %%
# Define a function to plot the different scores


def drawScore(score, tag):
    f = ot.DatabaseFunction(xLola, score)
    lb = distribution.getRange().getLowerBound()
    ub = distribution.getRange().getUpperBound()
    graph = f.draw(lb, ub)
    final = ot.Cloud(xLola)
    final.setPointStyle("fcircle")
    graph.add(final)
    graph.setTitle(f"{tag} score")
    graph.setXTitle("x1")
    graph.setYTitle("x2")
    otv.View(graph, square_axes=True)


# %%
# Plot the Voronoi score: it matches unexplored areas on top or right borders.
algo.generate(inc)  # triggers score update of the last batch
drawScore(algo.getVoronoiScore(), "Voronoi")


# %%
# Plot the LOLA score: it underlines regions with high gradients variations
drawScore(algo.getLOLAScore(), "LOLA")


# %%
# Plot the hybrid score: it exposes regions with medium-intensity gradients variations left to explore
drawScore(algo.getHybridScore(), "hybrid")


# %%
# Show all plots
otv.View.ShowAll()



#!/usr/bin/env python

import openturns as ot
import os

ot.TESTPREAMBLE()

process = ot.DiscreteMarkovChain()
print("Default constructor : process = ")
print(process)

process = ot.DiscreteMarkovChain(0, ot.SquareMatrix([[1.0, 0.0], [0.0, 1.0]]))
print("Constructor from int and SquareMatrix: process = ")
print(process)

transitionMatrix = ot.SquareMatrix([[0.0, 0.5, 0.5], [0.7, 0.0, 0.3], [0.8, 0.0, 0.2]])
print("transition matrix =")
print(transitionMatrix)

origin = 1
print("origin =")
print(origin)

process.setTransitionMatrix(transitionMatrix)
print("Transition matrix accessor : process = ")
print(process)

process.setOrigin(origin)
print("Origin accessor : process = ")
print(process)

process.setTimeGrid(ot.RegularGrid(0, 1, 20))
print("Time grid accessor : process = ")
print(process)

real = process.getRealization()
print("Realization :")
print(real)

future = process.getFuture(20)
print("One future :")
print(future)

futures = process.getFuture(20, 3)
print("3 different futures :")
print(futures)

dist = process.computeStationaryDistribution()
print("Distribution: ", dist)

process.exportToDOTFile("markov.dot")
with open("markov.dot") as f:
    print(f.read())
if 0:
    from graphviz import Source

    path = "markov.dot"
    s = Source.from_file(path)
    s.view()
os.remove("markov.dot")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)

size = 100
dim = 10
R = ot.CorrelationMatrix(dim)
for i in range(dim):
    for j in range(i):
        R[i, j] = (i + j + 1.0) / (2.0 * dim)

mean = [2.0] * dim
sigma = [3.0] * dim
distribution = ot.Normal(mean, sigma, R)

sample = distribution.getSample(size)
sampleX = sample.getMarginal(range(1, dim))
sampleY = sample.getMarginal(0)

sampleZ = ot.Sample(size, 1)
for i in range(size):
    sampleZ[i, 0] = sampleY[i, 0] ** 2
print(
    "LinearModelFisher pvalue=%1.2g"
    % ot.LinearModelTest.LinearModelFisher(sampleY, sampleZ).getPValue()
)
print(
    "LinearModelResidualMean pvalue=%1.2g"
    % ot.LinearModelTest.LinearModelResidualMean(sampleY, sampleZ).getPValue()
)

# Durbin Watson
ot.RandomGenerator.SetSeed(5415)
eps = ot.Normal(0, 20)
f = ot.SymbolicFunction("x", "5+2*x+x^2-0.1*x^3")
N = 15
x = ot.Sample(
    [
        [0],
        [1.42857],
        [2.85714],
        [4.28571],
        [5.71429],
        [7.14286],
        [8.57143],
        [10],
        [11.4286],
        [12.8571],
        [14.2857],
        [15.7143],
        [17.1429],
        [18.5714],
        [20],
    ]
)
y = f(x) + eps.getSample(N)
linmodel = ot.LinearModelAlgorithm(x, y).getResult().getCoefficients()
dwTest = ot.LinearModelTest.LinearModelDurbinWatson(x, y)
print("Durbin Watson = ", dwTest)


selection = ot.Indices(5)
selection.fill()

selection2 = ot.Indices(1, 0)
sampleX0 = sampleX.getMarginal(0)

# Regression test between 2 samples : firstSample of dimension n and
# secondSample of dimension 1. If firstSample[i] is the numerical sample
# extracted from firstSample (ith coordinate of each point of the
# numerical sample), PartialRegression performs the Regression test
# simultaneously on all firstSample[i] and secondSample, for i in the
# selection. The Regression test tests ifthe regression model between two
# scalar numerical samples is significant. It is based on the deviation
# analysis of the regression. The t-test is used.

# The two tests must be equal
print(
    "PartialRegressionX0Y=",
    ot.LinearModelTest.PartialRegression(sampleX, sampleY, selection2, 0.10),
)
print("FullRegressionX0Y=", ot.LinearModelTest.FullRegression(sampleX0, sampleY, 0.10))

print(
    "PartialRegressionXY=",
    ot.LinearModelTest.PartialRegression(sampleX, sampleY, selection, 0.10),
)

# Regression test between 2 samples : firstSample of dimension n and
# secondSample of dimension 1. If firstSample[i] is the numerical sample
# extracted from firstSample (ith coordinate of each point of the
# numerical sample), FullRegression performs the Regression test
# simultaneously on all firstSample[i] and secondSample. The Regression
# test tests if the regression model between two scalar numerical samples
# is significant. It is based on the deviation analysis of the regression.
# The t-test is used.

print("FullRegressionXZ=", ot.LinearModelTest.FullRegression(sampleX, sampleY, 0.10))
# print('FullRegressionZZ=', ot.LinearModelTest.FullRegression(
# sampleZ, sampleZ, 0.10))



"""
Gaussian Process Regression: choose a polynomial trend on the beam model
========================================================================
"""

# %%
# In this example, we consider the :ref:`cantilever beam <use-case-cantilever-beam>` and we build a metamodel using a
# Gaussian process regression (refer to :ref:`gaussian_process_regression`) whose trend is estimated on the
# given data set. We illustrate the impact of the choice of the trend function basis on the metamodel.
# This example focuses on three polynomial trends:
#
# * :class:`~openturns.ConstantBasisFactory`;
# * :class:`~openturns.LinearBasisFactory`;
# * :class:`~openturns.QuadraticBasisFactory`.
#
# In the :doc:`/auto_surrogate_modeling/gaussian_process_regression/plot_gpr_choose_trend` example,
# we give another example of this procedure.
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv


# %%
# Definition of the model
# -----------------------
#
# We load the use case.
cb = cantilever_beam.CantileverBeam()

# %%
# We define the function which evaluates the output depending on the inputs.
model = cb.model

# %%
# Then we define the distribution of the input random vector.
dimension = cb.dim  # number of inputs
input_dist = cb.distribution

# %%
# Create the design of experiments
# --------------------------------
#
# We consider a simple Monte-Carlo sampling as a design of experiments.
# This is why we generate an input sample using the :meth:`~openturns.Distribution.getSample` method of the distribution.
# Then we evaluate the output using the `model` function.

# %%
sampleSize_train = 20
X_train = input_dist.getSample(sampleSize_train)
Y_train = model(X_train)

# %%
# Constant basis
# --------------
#
# In this paragraph we choose a basis constant for the estimation of the trend.
# The basis is built with the :class:`~openturns.ConstantBasisFactory` class.
basis = ot.ConstantBasisFactory(dimension).build()

# %%
# In order to create the Gaussian Process Regression metamodel, we use a squared exponential covariance kernel.
# The :class:`~openturns.SquaredExponential` kernel has one magnitude coefficient and 4 scale coefficients.
# This is because this covariance kernel is anisotropic : each of the 4 input variables is associated with its own
# scale coefficient. We need to set the initial scale parameter for the optimization.
covariance_model = ot.SquaredExponential(dimension)

# %%
# To create the Gaussian Process Regression metamodel, we first build the :math:`Y(\omega, x)` Gaussian process with the class
# :class:`~openturns.GaussianProcessFitter`. It requires a training sample, a covariance kernel and a
# trend basis as input arguments.
algo_fit = ot.GaussianProcessFitter(X_train, Y_train, covariance_model, basis)
algo_fit.run()
fit_result = algo_fit.getResult()

# %%
# Then, we condition the process :math:`Y(\omega, x)` to the data set with the class
# :class:`~openturns.GaussianProcessRegression`.
algo_gpr = ot.GaussianProcessRegression(fit_result)
algo_gpr.run()

# %%
# Get the Gaussian Process Regression metamodel.
gpr_result_cst = algo_gpr.getResult()
metamodel_cst = gpr_result_cst.getMetaModel()

# %%
# The :meth:`~openturns.GaussianProcessRegressionResult.getTrendCoefficients` method returns the coefficients of the trend.
# The constant trend always has only one coefficient (if there is one single output).
print(gpr_result_cst.getTrendCoefficients())

# %%
# We can check the estimated covariance model.
print(gpr_result_cst.getCovarianceModel())

# %%
# Linear basis
# ------------
#
# In this paragraph we choose a linear basis for the estimation of the trend.
# The basis is built with the :class:`~openturns.LinearBasisFactory` class. The same methodology is followed:
# we do not detail it.
basis = ot.LinearBasisFactory(dimension).build()
algo_fit = ot.GaussianProcessFitter(X_train, Y_train, covariance_model, basis)
algo_fit.run()
fit_result = algo_fit.getResult()
algo_gpr = ot.GaussianProcessRegression(fit_result)
algo_gpr.run()
gpr_result_lin = algo_gpr.getResult()
metamodel_lin = gpr_result_lin.getMetaModel()
print(gpr_result_lin.getTrendCoefficients())
print(gpr_result_lin.getCovarianceModel())

# %%
# The number of coefficients in the linear and quadratic trends depends on the number of inputs, which is
# equal to
#
# .. math::
#    dim = 4
#
#
# in the cantilever beam case.
#
# We observe that the number of coefficients in the trend is 5, which corresponds to:
#
# * 1 coefficient for the constant part,
# * dim = 4 coefficients for the linear part.
#
#
# Quadratic basis
# ---------------
#
# In this paragraph we choose a quadratic basis for the estimation of the trend.
# The basis is built with the :class:`~openturns.QuadraticBasisFactory` class. The same methodology is followed:
# we do not detail it.
#
# However we can see that the default optimization algorithm which is :class:`~openturns.Cobyla` does not manage to converge.
# Thus, we can either:
#
# * change the default optimization algorithm and select for example the :class:`~openturns.TNC`
#   algorithm (Truncated Newton Constrained) using the entry of :class:`~openturns.ResourceMap` called
#   *GaussianProcessFitter-DefaultOptimizationAlgorithm*: *ot.ResourceMap.SetAsString("GaussianProcessFitter-
#   DefaultOptimizationAlgorithm", "TNC")*,
# * or keep the default optimization algorithm but change the default maximum constrainte error value which is equal to
#   :math:`10^{-5}`. We move it to :math:`10^{-6}`  using the entry of :class:`~openturns.ResourceMap` called
#   *OptimizationAlgorithm-DefaultMaximumConstraintError*: *ot.ResourceMap.SetAsScalar("OptimizationAlgorithm
#   -DefaultMaximumConstraintError", 1e-6)*.
#
ot.ResourceMap.SetAsScalar("OptimizationAlgorithm-DefaultMaximumConstraintError", 1e-6)
basis = ot.QuadraticBasisFactory(dimension).build()
algo_fit = ot.GaussianProcessFitter(X_train, Y_train, covariance_model, basis)
algo_fit.run()
fit_result = algo_fit.getResult()
algo_gpr = ot.GaussianProcessRegression(fit_result)
algo_gpr.run()
gpr_result_quad = algo_gpr.getResult()
metamodel_quad = gpr_result_quad.getMetaModel()
print(gpr_result_quad.getTrendCoefficients())
print(gpr_result_quad.getCovarianceModel())

# %%
# This time, the trend has 15 coefficients:
#
# * 1 coefficient for the constant part,
# * 4 coefficients for the linear part,
# * 10 coefficients for the quadratic part.
#
# This is because the number of coefficients in the quadratic part has
#
# .. math::
#    \frac{dim \times (dim+1)}{2}=\frac{4\times 5}{2}=10
#
#
# coefficients, associated with the symmetric matrix of the quadratic function.
#
# Validate the metamodel
# ----------------------
#
# We finally want to validate the Gaussian Process Regression metamodel. This is why we generate a validation sample
# with size 100 and we evaluate the output of the model on this sample.
sampleSize_test = 100
X_test = input_dist.getSample(sampleSize_test)
Y_test = model(X_test)

# %%
# We define a function to easily draw the QQ-plot graphs.


def drawMetaModelValidation(X_test, Y_test, metamodel_gpr, title):
    metamodel_predictions = metamodel_gpr(X_test)
    val = ot.MetaModelValidation(Y_test, metamodel_predictions)
    r2Score = val.computeR2Score()[0]
    graph = val.drawValidation().getGraph(0, 0)
    graph.setLegends([""])
    graph.setLegends(["%s, R2 = %.2f%%" % (title, 100 * r2Score), ""])
    graph.setLegendPosition("upper left")
    return graph


# %%
# We plot here the validation graph for each metamodel.
grid = ot.GridLayout(1, 3)
grid.setTitle("Different trends")
graphConstant = drawMetaModelValidation(X_test, Y_test, metamodel_cst, "Constant")
graphLinear = drawMetaModelValidation(X_test, Y_test, metamodel_lin, "Linear")
graphQuadratic = drawMetaModelValidation(X_test, Y_test, metamodel_quad, "Quadratic")
grid.setGraph(0, 0, graphConstant)
grid.setGraph(0, 1, graphLinear)
grid.setGraph(0, 2, graphQuadratic)
view = otv.View(grid)

# %%
# We observe that the three trends perform very well in this case.
# With more coefficients, the Gaussian Process Regression metamodel is more flexibile and can adjust better to the training sample.
# This does not mean, however, that the trend coefficients will provide a good fit for the validation sample.
#
# The number of parameters in each Gaussian Process Regression metamodel is the following :
#
# * with the constant trend, we have 6 coefficients to estimate: 5 coefficients for the covariance matrix and 1 coefficient for the trend,
# * with the  linear trend, we have 10 coefficients to estimate: 5 coefficients for the covariance matrix and 5 coefficients for the trend,
# * with the quadratic trend, we have 20 coefficients to estimate: 5 coefficients for the covariance matrix and 15 coefficients for the trend.
#
# In the cantilever beam example, fitting the metamodel to a training sample with only 10 points is made much easier because the function to approximate is almost linear.
# In this case, a quadratic trend is not advisable because it can interpolate all points in the training sample.


# %%
# Display figures
otv.View.ShowAll()



"""
Create a Python function
========================
"""

# %%
# In this example we are going to create a :class:`~openturns.Function` object (ie usable throughout the library) from a pure Python function.
#
# The pure Python function to wrap must accept a sequence of floats and return a sequence of float.
#
# .. math::
#    f(x) = [x_0+x_1+x_2, (x_1-1)*e^{x_0} * x_2]
#

# %%
import numpy as np
import openturns as ot
import math as m

# %%
# define a pure Python function from R^3 to R^2


def regularFunc(X):
    x0, x1, x2 = X
    y0 = x0 + x1 + x2
    y1 = (x1 - 1.0) * m.exp(x0) * x2
    return [y0, y1]


# %%
# create a Function object from a regular Python function
function = ot.PythonFunction(3, 2, regularFunc)

# %%
# evaluate the function on a Point
x = [1.0, 2.0, 3.0]
print("x=", x, "f(x)=", function(x))

# %%
# evaluate the function on a Sample
xs = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]
print("xs=", xs, "\nf(xs)=", function(xs))

# %%
# now we can use the Function object services such as the gradient
function.gradient(x)

# %%
# Performance issues
# ------------------
#
# When this function is used internally to evaluate a Sample, it loops over its points.
# This requires many memory allocations; moreover this loop is done in Python, it may thus be slow if Sample is large.
# We can define a function to operate on a Sample, and return a Sample.
#
# For maximum performance, argument is in fact not a Sample, but a wrapper
# object which contains a pointer to data.  When using Numpy arrays without
# copies and loops, performance is similar to C code, but Python definition
# is somewhat convoluted; please refer to Numpy documentation to learn how to efficiently define such functions.

# %%
# define the same function on a Sample


def regularFuncSample(X):
    # Create a numpy array with the contents of X without copy
    xarray = np.asarray(X)
    # Get columns as vectors, there is also no copy
    x0, x1, x2 = xarray.T
    # Allocate a numpy array to store result
    y = np.zeros((len(X), 2))
    y[:, 0] = x0 + x1 + x2
    y[:, 1] = (x1 - 1.0) * np.exp(x0) * x2
    return y


# %%
# create a Function object from a regular Python function
functionSample = ot.PythonFunction(3, 2, func_sample=regularFuncSample)

# %%
# evaluate the function on a Sample
print("xs=", xs, "\nf(xs)=", functionSample(xs))

# %%
# evaluate the function on a Point
print("x=", x, "f(x)=", functionSample(x))

# %%
# The most efficient solution is to provide evaluations both on Point and Sample.
# This requires two Python function definitions, but if your code takes a lot of time, you should consider this option.

# %%
functionFast = ot.PythonFunction(3, 2, func=regularFunc, func_sample=regularFuncSample)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

for dim in [2, 5, 10]:
    print("dimension = ", dim)

    # We create a numerical math function
    inputVar = ["X" + str(i) for i in range(dim)]
    myFunction = ot.SymbolicFunction(inputVar, ["0"])

    # We create a normal distribution point of dimension 1
    mean = [0.0] * dim
    sigma = [1.0] * dim
    R = ot.IdentityMatrix(dim)
    myDistribution = ot.Normal(mean, sigma, R)

    # We create a 'usual' RandomVector from the Distribution
    vect = ot.RandomVector(myDistribution)

    # We create a composite random vector
    output = ot.CompositeRandomVector(myFunction, vect)

    # We create a StandardEvent from this RandomVector
    myStandardEvent = ot.StandardEvent(output, ot.Less(), 2.0)

    std = ot.Normal()

    beta = ot.Point(3)
    beta[0] = round(-std.computeQuantile(1e-3)[0])
    beta[1] = round(-std.computeQuantile(1e-5)[0])
    beta[2] = round(-std.computeQuantile(1e-7)[0])

    importanceLevel = ot.Point(3)
    importanceLevel[0] = 0.01
    importanceLevel[1] = 0.05
    importanceLevel[2] = 0.10

    accuracyLevel = ot.Point(3)
    accuracyLevel[0] = 1.5
    accuracyLevel[1] = 2.0
    accuracyLevel[2] = 4.0

    confidenceLevel = ot.Point(3)
    confidenceLevel[0] = 0.90
    confidenceLevel[1] = 0.95
    confidenceLevel[2] = 0.99

    pointNumber = ot.UnsignedIntegerCollection(3)
    pointNumber[0] = 10
    pointNumber[1] = 100
    pointNumber[2] = 1000

    # TABLE 1 : we impose beta, the importance level, the accuracy level,
    # tne confidence level and we calculate the corresponding deltaEpsilon
    # and pointNumber N
    print(
        "beta ",
        "importanceLevel ",
        "accuracyLevel ",
        "confidenceLevel ",
        "deltaEpsilon ",
        "pointNumber",
    )

    # loop on beta
    for indexBeta in range(beta.getDimension()):
        # We create the design point
        designPoint = ot.Point(dim, 0.0)
        designPoint[0] = beta[indexBeta]

        # loop on the importance level epsilon
        for indexImportanceLevel in range(importanceLevel.getDimension()):
            # loop on the accuracy level tau
            for indexAccuracyLevel in range(accuracyLevel.getDimension()):
                # loop on the confidence level (1-q)
                for indexConfidenceLevel in range(confidenceLevel.getDimension()):
                    # we calculate the corresponding deltaEpsilon and
                    # pointNumber N
                    myTest = ot.StrongMaximumTest(
                        myStandardEvent,
                        designPoint,
                        importanceLevel[indexImportanceLevel],
                        accuracyLevel[indexAccuracyLevel],
                        confidenceLevel[indexConfidenceLevel],
                    )
                    print(
                        "%.6f" % beta[indexBeta],
                        " %.6f" % importanceLevel[indexImportanceLevel],
                        " %.6f" % accuracyLevel[indexAccuracyLevel],
                        " %.6f" % confidenceLevel[indexConfidenceLevel],
                        " %.6f" % myTest.getDeltaEpsilon(),
                        " %.6f" % myTest.getPointNumber(),
                    )

                    # TABLE 2 : we impose beta, the importance level, the accuracy level, the pointNumber N  and we calculate the corresponding deltaEpsilon and confidence level
                    #       std::cout << std::right
                    #                 << std::setw(10) << "beta "
                    #                 << std::setw(16) << "importanceLevel "
                    #                 <<  "accuracyLevel " << "pointNumber " << "deltaEpsilon " << "confidenceLevel" << std::endl
    print(
        "beta ",
        "importanceLevel ",
        "accuracyLevel ",
        "pointNumber",
        "deltaEpsilon ",
        "confidenceLevel",
    )

    # loop on beta
    for indexBeta in range(beta.getDimension()):
        # We create the design point
        designPoint = ot.Point(dim, 0.0)
        designPoint[0] = beta[indexBeta]

        # loop on the importance level epsilon
        for indexImportanceLevel in range(importanceLevel.getDimension()):
            # loop on the accuracy level tau
            for indexAccuracyLevel in range(accuracyLevel.getDimension()):
                # loop on the pointNumber N
                for indexPointNumber in range(pointNumber.getSize()):
                    # we calculate the corresponding deltaEpsilon and
                    # confidenceLevel
                    myTest = ot.StrongMaximumTest(
                        myStandardEvent,
                        designPoint,
                        importanceLevel[indexImportanceLevel],
                        accuracyLevel[indexAccuracyLevel],
                        pointNumber[indexPointNumber],
                    )
                    print(
                        "%.6f" % beta[indexBeta],
                        " %.6f" % importanceLevel[indexImportanceLevel],
                        " %.6f" % accuracyLevel[indexAccuracyLevel],
                        " %.6f" % pointNumber[indexPointNumber],
                        " %.6f" % myTest.getDeltaEpsilon(),
                        " %.6f" % myTest.getConfidenceLevel(),
                    )



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.GumbelCopula(1.5)
size = 1000
sample = distribution.getSample(size)
factory = ot.GumbelCopulaFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedGumbelCopula = factory.buildAsGumbelCopula(sample)
print("GumbelCopula          =", distribution)
print("Estimated gumbelCopula=", estimatedGumbelCopula)
estimatedGumbelCopula = factory.buildAsGumbelCopula()
print("Default gumbelCopula=", estimatedGumbelCopula)
estimatedGumbelCopula = factory.buildAsGumbelCopula(distribution.getParameter())
print("GumbelCopula from parameters=", estimatedGumbelCopula)



"""
Transform a distribution
========================
"""

# %%
# In this example we are going to use distribution algebra and distribution transformation via functions.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We define some (classical) distributions
distribution1 = ot.Uniform(0.0, 1.0)
distribution2 = ot.Uniform(0.0, 2.0)
distribution3 = ot.WeibullMin(1.5, 2.0)

# %%
# Sum & difference of distributions
# ---------------------------------
#
# It is easy to compute the sum of distributions. For example:

# %%
distribution = distribution1 + distribution2
print(distribution)
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# We might also use subtraction even with scalar values:

# %%
distribution = 3.0 - distribution3
print(distribution)
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Product & inverse
# -----------------
#
# We might also compute the product of two (or more) distributions. For example:

# %%
distribution = distribution1 * distribution2
print(distribution)
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# We could also inverse a distribution :

# %%
distribution = 1 / distribution1
print(distribution)
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Or compute a ratio distribution :

# %%
ratio = distribution2 / distribution1
print(ratio)
graph = ratio.drawPDF()
view = otv.View(graph)

# %%
# Transformation using functions
# ------------------------------
#
# The library provides methods to get the full distributions of `f(x)` where `f` can be equal to:
#
#  - `sin`,
#  - `asin`,
#  - `cos`,
#  - `acos`,
#  - `tan`,
#  - `atan`,
#  - `sinh`,
#  - `asinh`,
#  - `cosh`,
#  - `acosh`,
#  - `tanh`,
#  - `atanh`,
#  - `sqr` (for square),
#  - `inverse`,
#  - `sqrt`,
#  - `exp`,
#  - `log`/`ln`,
#  - `abs`,
#  - `cbrt`.
#
#
# For example for the usual `log` transformation:

# %%
graph = distribution1.log().drawPDF()
view = otv.View(graph)

# %%
# And for the `log2` function :

# %%
f = ot.SymbolicFunction(["x"], ["log2(x)"])
f.setDescription(["X", "ln(X)"])
graph = ot.CompositeDistribution(f, distribution1).drawPDF()
view = otv.View(graph)

# %%
# If one wants a specific method, user might rely on the :class:`~openturns.CompositeDistribution` class.

# Create a composite distribution
# -------------------------------
#
# In this paragraph we create a distribution defined as the push-forward distribution of a scalar distribution by a transformation.
#
# If we note :math:`\mathcal{L}_0` a scalar distribution, :math:`f: \mathbb{R} \rightarrow \mathbb{R}` a mapping,
# then it is possible to create the push-forward distribution :math:`\mathcal{L}` defined by
#
# .. math::
#    \mathcal{L} = f(\mathcal{L}_0)
#

# %%
# We create a 1-d Normal distribution
antecedent = ot.Normal()

# %%
# and a 1-d transform :
f = ot.SymbolicFunction(["x"], ["sin(x)+cos(x)"])

# %%
# We then create the composite distribution
distribution = ot.CompositeDistribution(f, antecedent)
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# We can also build a distribution with the simplified construction
distribution = antecedent.exp()
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# and by using chained operators:
distribution = antecedent.abs().sqrt()
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Create a PCE by integration on the cantilever beam
==================================================
"""

# %%
# In this example, we create a polynomial chaos metamodel by integration on the
# :ref:`cantilever beam <use-case-cantilever-beam>` example.
# We choose to evaluate the coefficients of the chaos decomposition by
# integration using various kinds of design of experiments:
#
# - Gauss product,
# - Latin hypercube sampling,
# - Quasi Monte-Carlo with a Sobol' sequence.
#
# We will compare the results obtained on each design.

# %%
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv


# %%
# We first load the model from the usecases module :
cb = cantilever_beam.CantileverBeam()

# %%
# In this example we consider all marginals independent.
# They are defined in the :class:`~openturns.usecases.cantilever_beam.CantileverBeam`  class:
dist_E = cb.E
dist_F = cb.F
dist_L = cb.L
dist_I = cb.II
distribution = cb.independentDistribution

# %%
# We load the model.
dim_input = cb.dim  # dimension of the input
dim_output = 1  # dimension of the output
g = cb.model

# %%
# Create a polynomial chaos decomposition
# ---------------------------------------

# %%
# We create the multivariate polynomial basis by tensorization of the
# univariate polynomials and the default linear enumerate rule.
multivariateBasis = ot.OrthogonalProductPolynomialFactory(
    [dist_E, dist_F, dist_L, dist_I]
)

# %%
# In this case, we select :math:`P` using the
# :meth:`~openturns.EnumerateFunction.getBasisSizeFromTotalDegree` method,
# so that all polynomials with total degree lower or equal to 5 are used.
# This will lead to the computation of 126 coefficients.
totalDegree = 5
enum_func = multivariateBasis.getEnumerateFunction()
basisSize = enum_func.getBasisSizeFromTotalDegree(totalDegree)
print(f"Basis size = {basisSize}")

# %%
# We select the :class:`~openturns.FixedStrategy` truncation rule, which corresponds to using
# the first :math:`P` polynomials of the polynomial basis.
adaptiveStrategy = ot.FixedStrategy(multivariateBasis, basisSize)

# %%
# We begin by getting the standard measure associated with the multivariate polynomial basis.
# We see that the range of the `Beta` distribution has been standardized into the `[-1, 1]` interval.
# This is the same for the `Uniform` distribution and the second `Beta` distribution.
measure = multivariateBasis.getMeasure()
print(f"Measure = {measure}")

# %%
# The choice of the :class:`~openturns.GaussProductExperiment` rule with 4 nodes
# in each of the 4 dimensions leads to :math:`4^4=256` evaluations of the model.
marginalSizes = [4] * dim_input
experiment = ot.GaussProductExperiment(distribution, marginalSizes)
print(f"N={experiment.getSize()}")
X, W = experiment.generateWithWeights()
Y = g(X)

# %%
# We now set the method used to compute the coefficients; we select the integration method.
projectionStrategy = ot.IntegrationStrategy()

# %%
# We can now create the functional chaos.
algo = ot.FunctionalChaosAlgorithm(
    X, W, Y, distribution, adaptiveStrategy, projectionStrategy
)
algo.run()

# %%
# Get the result
result = algo.getResult()

# %%
# The :meth:`~openturns.FunctionalChaosResult.getMetaModel` method returns the metamodel function.
metamodel = result.getMetaModel()

# %%
# Validate the metamodel
# ----------------------

# %%
# Generate a new validation sample (which is independent of the training sample).
n_valid = 1000
X_test = distribution.getSample(n_valid)
Y_test = g(X_test)

# %%
# The :class:`~openturns.MetaModelValidation` class validates the metamodel
# based on a validation sample.
metamodelPredictions = metamodel(X_test)
val = ot.MetaModelValidation(Y_test, metamodelPredictions)

# %%
# Compute the :math:`R^2` coefficient of determination.
r2Score = val.computeR2Score()[0]
r2Score

# %%
# Plot the observed versus the predicted outputs.
graph = val.drawValidation()
graph.setTitle(f"Gauss product N={experiment.getSize()} - R2={r2Score * 100:.2f}")
view = otv.View(graph)

# %%
# Now repeat the same process on various designs.


def draw_validation(experiment):
    projectionStrategy = ot.IntegrationStrategy(experiment)
    algo = ot.FunctionalChaosAlgorithm(
        X, Y, distribution, adaptiveStrategy, projectionStrategy
    )
    algo.run()
    result = algo.getResult()
    metamodel = result.getMetaModel()
    X_test = distribution.getSample(n_valid)
    Y_test = g(X_test)
    metamodelPredictions = metamodel(X_test)
    val = ot.MetaModelValidation(Y_test, metamodelPredictions)
    r2Score = val.computeR2Score()[0]
    graph = val.drawValidation()
    graph.setTitle(
        f"{experiment.__class__.__name__} - N={experiment.getSize()} - R2={r2Score * 100:.2f}"
    )
    return graph


# %%
# Use an LHS design.
experiment = ot.LHSExperiment(distribution, int(1e6))
graph = draw_validation(experiment)
view = otv.View(graph)

# %%
# Use a low-discrepancy experiment (Quasi-Monte Carlo).
sequence = ot.SobolSequence()
experiment = ot.LowDiscrepancyExperiment(sequence, distribution, int(1e5))
graph = draw_validation(experiment)
view = otv.View(graph)

otv.View.ShowAll()

# %%
# Conclusion
# ----------
# With the Gauss product rule, the coefficients are particularly well computed
# since the R2 score is excellent, even with the relatively limited amount
# of simulation (256 points).
# On the other hand the LHS and low-discrepancy experiments require many more
# points to achieve a :math:`R^2` > 99%.



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott


ot.TESTPREAMBLE()

factories = ot.DistributionFactory.GetUniVariateFactories()
for factory in factories:
    dist = factory.build()
    with ott.assert_raises(TypeError):
        qm1 = dist.computeQuantile(-1.0)
    with ott.assert_raises(TypeError):
        q2 = dist.computeQuantile(2.0)
    q0 = dist.computeQuantile(0.0)[0]
    q1 = dist.computeQuantile(1.0)[0]
    q0p = dist.computeQuantile(ot.SpecFunc.MinScalar)[0]
    if not q0 <= q0p:
        print(dist.getName(), "0+", q0, q0p)
    q1 = dist.computeQuantile(1.0)[0]
    q1m = dist.computeQuantile(1.0 - ot.SpecFunc.ScalarEpsilon)[0]
    if not q1m <= q1:
        print(dist.getName(), "1-", q1m, q1)



"""
Create an aggregated function
=============================
"""

# %%
# In this example we are going to build a function that stacks all the outputs from several functions
#
# .. math::
#    f = (f_1, \dots, f_n)
#

# %%
import openturns as ot


# %%
# assume a list of functions to aggregate
functions = list()
functions.append(ot.SymbolicFunction(["x1", "x2", "x3"], ["x1^2 + x2", "x1 + x2 + x3"]))
functions.append(
    ot.SymbolicFunction(["x1", "x2", "x3"], ["x1 + 2 * x2 + x3", "x1 + x2 - x3"])
)

# %%
# create the aggregated function
function = ot.AggregatedFunction(functions)

# %%
# evaluate the function
x = [1.0, 2.0, 3.0]
y = function(x)
print("x=", x, "y=", y)



#!/usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# create a function
dim = 4
function = ot.SymbolicFunction(["E", "F", "L", "I"], ["F*L^3/(3.*E*I)"])

# create a distribution
distribution = ot.Normal([50.0, 1.0, 10.0, 5.0], [1.0] * dim, ot.IdentityMatrix(dim))
vect = ot.RandomVector(distribution)
composite = ot.CompositeRandomVector(function, vect)
event = ot.ThresholdEvent(composite, ot.Less(), -3.0)

# create an ADS algorithm
n = int(1e4)
algo = ot.AdaptiveDirectionalStratification(event)
algo.setMaximumOuterSampling(n)
algo.setGamma([0.6, 0.4])

algo.run()
result = algo.getResult()
print(result)

# ADS-2+
algo2 = algo
algo2.setPartialStratification(True)
algo2.run()
print("T=", algo2.getTStatistic())
result = algo2.getResult()
print(result)

# DPADS-2
algo3 = algo2
algo3.setQuadrantOrientation([1.0] * dim)  # enables DPADS-2, sets design
algo3.run()
result = algo3.getResult()
print(result)



"""
Use case : logistic model
=========================
"""

import openturns as ot
import math


class LogisticModel:
    r"""
    Data class for the logistic model.

    In the physical model, the inputs and parameters are ordered as
    presented in the next table.
    Notice that there are no parameters in the physical model.

    +-------+----------------+
    | Index | Input variable |
    +=======+================+
    | 0     | t1             |
    +-------+----------------+
    | 1     | t2             |
    +-------+----------------+
    | ...   | ...            |
    +-------+----------------+
    | 21    | t22            |
    +-------+----------------+
    | 22    | a              |
    +-------+----------------+
    | 23    | c              |
    +-------+----------------+

    Attributes
    ----------

    t0 : float, optional
        Initial time. The default is 1790.

    y0 : float, optional
        Initial population (at t0). The default is 3.9e6.

    a : float, optional
        8Parameter of the model. The default is 0.03134.

    b : float, optional
        Parameter of the model. The default is 1.5887e-10.

    populationFactor : float, optional
        The multiplication factor to scale the population.
        The default is 1.0e6.

    distY0 : :class:`~openturns.Normal`
        `ot.Normal(y0, 0.1 * y0)`

    distA : :class:`~openturns.Normal`
        `ot.Normal(a, 0.3 * a)`

    distB : :class:`~openturns.Normal`
        `ot.Normal(b, 0.3 * b)`

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    model : :class:`~openturns.PythonFunction`
        The logistic model of growth.
        The input has input dimension 24 and output dimension 22.
        More precisely, we have :math:`\vect{X} = (t_1, ..., t_{22}, a, c)`
        and  :math:`\vect{Y} = (y_1, ..., y_{22})`.

    data : :class:`~openturns.Sample`
        A data set containing 22 dates from 1790 to 2000.
        First marginal represents dates and second marginal the population in millions.

    Examples
    --------
    >>> from openturns.usecases import logistic_model
    >>> # Load the logistic model
    >>> lm = logistic_model.LogisticModel()
    >>> print(lm.data[:5])
        [ Time            U.S. Population ]
    0 : [ 1790               3.9          ]
    1 : [ 1800               5.3          ]
    2 : [ 1810               7.2          ]
    3 : [ 1820               9.6          ]
    4 : [ 1830              13            ]
    >>> print("Inputs:", lm.model.getInputDescription())
    Inputs: [t0,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t13,t14,t15,t16,t17,t18,t19,t20,t21,a,c]#24
    >>> print("Outputs:", lm.model.getOutputDescription())
    Outputs: [z0,z1,z2,z3,z4,z5,z6,z7,z8,z9,z10,z11,z12,z13,z14,z15,z16,z17,z18,z19,z20,z21]#22
    """

    def __init__(
        self, t0=1790.0, y0=3.9e6, a=0.03134, b=1.5887e-10, populationFactor=1.0e6
    ):
        # Initial value of the population
        self.t0 = t0
        self.y0 = y0
        self.populationFactor = populationFactor
        self.a = a
        self.b = b

        self.distY0 = ot.Normal(self.y0, 0.1 * self.y0)
        self.distA = ot.Normal(self.a, 0.3 * self.a)
        self.distB = ot.Normal(self.b, 0.3 * self.b)
        self.distribution = ot.JointDistribution([self.distY0, self.distA, self.distB])
        self.inputDistribution = self.distribution  # deprecated

        # Observation points
        self.data = ot.Sample(
            [
                [1790, 3.9],
                [1800, 5.3],
                [1810, 7.2],
                [1820, 9.6],
                [1830, 13],
                [1840, 17],
                [1850, 23],
                [1860, 31],
                [1870, 39],
                [1880, 50],
                [1890, 62],
                [1900, 76],
                [1910, 92],
                [1920, 106],
                [1930, 123],
                [1940, 132],
                [1950, 151],
                [1960, 179],
                [1970, 203],
                [1980, 221],
                [1990, 250],
                [2000, 281],
            ]
        )
        self.data.setDescription(["Time", "U.S. Population"])
        nbdates = self.data.getSize()

        def logisticModel(X):
            t = [X[i] for i in range(nbdates)]
            a = X[22]
            c = X[23]
            b = math.exp(c)
            y = [0.0] * nbdates
            for i in range(nbdates):
                y[i] = a * y0 / (b * y0 + (a - b * y0) * math.exp(-a * (t[i] - t0)))
            z = [yi / populationFactor for yi in y]  # Convert into millions
            return z

        self.model = ot.PythonFunction(24, nbdates, logisticModel)
        inputLabels = [f"t{i}" for i in range(nbdates)]
        inputLabels.append("a")
        inputLabels.append("c")
        outputLabels = [f"z{i}" for i in range(nbdates)]
        self.model.setInputDescription(inputLabels)
        self.model.setOutputDescription(outputLabels)
